import logging
try:
    import OpenSSL.SSL
    import cryptography.hazmat.backends.openssl.backend
except ImportError:
    OPENSSL_AVAILABLE = False
else:
    OPENSSL_AVAILABLE = True

from . import tds_base


BUFSIZE = 65536


logger = logging.getLogger(__name__)


class EncryptedSocket(object):
    def __init__(self, transport, tls_conn):
        self._transport = transport
        self._tls_conn = tls_conn

    def gettimeout(self):
        return self._transport.gettimeout()

    def settimeout(self, timeout):
        self._transport.settimeout(timeout)

    def sendall(self, data, flags=0):
        # TLS.Connection does not support bytearrays, need to convert to bytes first
        if isinstance(data, bytearray):
            data = bytes(data)

        res = self._tls_conn.sendall(data)
        buf = self._tls_conn.bio_read(BUFSIZE)
        self._transport.sendall(buf)
        return res

 #   def send(self, data):
 #       while True:
 #           try:
 #               return self._tls_conn.send(data)
 #           except OpenSSL.SSL.WantWriteError:
 #               buf = self._tls_conn.bio_read(BUFSIZE)
 #               self._transport.sendall(buf)

    def recv_into(self, buffer, size=0):
        if size == 0:
            size = len(buffer)
        res = self.recv(size)
        buffer[0:len(res)] = res
        return len(res)

    def recv(self, bufsize):
        while True:
            try:
                buf = self._tls_conn.bio_read(bufsize)
            except OpenSSL.SSL.WantReadError:
                pass
            else:
                self._transport.sendall(buf)

            try:
                return self._tls_conn.recv(bufsize)
            except OpenSSL.SSL.WantReadError:
                buf = self._transport.recv(BUFSIZE)
                if buf:
                    self._tls_conn.bio_write(buf)
                else:
                    return b''

    def close(self):
        self._tls_conn.shutdown()
        self._transport.close()

    def shutdown(self):
        self._tls_conn.shutdown()


def verify_cb(conn, cert, err_num, err_depth, ret_code):
    return ret_code == 1


def validate_host(cert, name):
    """
    Validates host name against certificate

    @param cert: Certificate returned by host
    @param name: Actual host name used for connection
    @return: Returns true if host name matches certificate
    """
    cn = None
    for t, v in cert.get_subject().get_components():
        if t == b'CN':
            cn = v
            break

    if cn == name:
        return True

    # checking SAN
    s_name = name.decode('ascii')
    for i in range(cert.get_extension_count()):
        ext = cert.get_extension(i)
        if ext.get_short_name() == b'subjectAltName':
            s = str(ext)
            # SANs are usually have form like: DNS:hostname
            if s.startswith('DNS:') and s[4:] == s_name:
                return True

    # TODO handle wildcards
    return False


def create_context(cafile):
    ctx = OpenSSL.SSL.Context(OpenSSL.SSL.TLSv1_2_METHOD)
    ctx.set_options(OpenSSL.SSL.OP_NO_SSLv2)
    ctx.set_options(OpenSSL.SSL.OP_NO_SSLv3)
    ctx.set_verify(OpenSSL.SSL.VERIFY_PEER, verify_cb)
    #print("verify depth:", ctx.get_verify_depth())
    #print("verify mode:", ctx.get_verify_mode())
    #print("openssl version:", cryptography.hazmat.backends.openssl.backend.openssl_version_text())
    ctx.load_verify_locations(cafile=cafile)
    return ctx


# https://msdn.microsoft.com/en-us/library/dd357559.aspx
def establish_channel(tds_sock):
    w = tds_sock._writer
    r = tds_sock._reader
    login = tds_sock.conn._login

    bhost = login.server_name.encode('ascii')

    conn = OpenSSL.SSL.Connection(login.tls_ctx)
    conn.set_tlsext_host_name(bhost)
    # change connection to client mode
    conn.set_connect_state()
    logger.info('doing TLS handshake')
    while True:
        try:
            logger.debug('calling do_handshake')
            conn.do_handshake()
        except OpenSSL.SSL.WantReadError:
            logger.debug('got WantReadError, getting data from the write end of the TLS connection buffer')
            try:
                req = conn.bio_read(BUFSIZE)
            except OpenSSL.SSL.WantReadError:
                # PyOpenSSL - https://github.com/pyca/pyopenssl/issues/887
                logger.debug('got WantReadError again, waiting for response...')
            else:
                logger.debug('sending %d bytes of the handshake data to the server', len(req))
                w.begin_packet(tds_base.PacketType.PRELOGIN)
                w.write(req)
                w.flush()
            logger.debug('receiving response from the server')
            resp = r.read_whole_packet()
            # TODO validate r.packet_type
            logger.debug('adding %d bytes of the response into the TLS connection buffer', len(resp))
            conn.bio_write(resp)
        else:
            logger.info('TLS handshake is complete')
            if login.validate_host:
                if not validate_host(cert=conn.get_peer_certificate(), name=bhost):
                    raise tds_base.Error("Certificate does not match host name '{}'".format(login.server_name))
            enc_sock = EncryptedSocket(transport=tds_sock.conn.sock, tls_conn=conn)
            tds_sock.conn.sock = enc_sock
            tds_sock._writer._transport = enc_sock
            tds_sock._reader._transport = enc_sock
            return


def revert_to_clear(tds_sock):
    """
    Reverts connection back to non-encrypted mode
    Used when client sent ENCRYPT_OFF flag
    @param tds_sock:
    @return:
    """
    enc_conn = tds_sock.conn.sock
    clear_conn = enc_conn._transport
    enc_conn.shutdown()
    tds_sock.conn.sock = clear_conn
    tds_sock._writer._transport = clear_conn
    tds_sock._reader._transport = clear_conn



import os.path
import shutil
import datetime
import pathlib
import cryptography.hazmat.backends
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography import x509
import cryptography.x509.oid


class TestCA:
    def __init__(self):
        self._key_cache = {}
        backend = cryptography.hazmat.backends.default_backend()
        self._test_cache_dir = os.path.join(os.path.dirname(__file__), '..', '.test-cache')
        os.makedirs(self._test_cache_dir, exist_ok=True)
        root_cert_path = self.cert_path('root')
        self._root_key = self.key('root')
        self._root_ca = generate_root_certificate(self._root_key)
        pathlib.Path(root_cert_path).write_bytes(self._root_ca.public_bytes(serialization.Encoding.PEM))

    def key_path(self, name):
        return os.path.join(self._test_cache_dir, name + 'key.pem')

    def cert_path(self, name):
        return os.path.join(self._test_cache_dir, name + 'cert.pem')

    def key(self, name) -> rsa.RSAPrivateKey:
        if name not in self._key_cache:
            backend = cryptography.hazmat.backends.default_backend()
            key_path = self.key_path(name)
            if os.path.exists(key_path):
                bin = pathlib.Path(key_path).read_bytes()
                key = serialization.load_pem_private_key(bin, password=None, backend=backend)
            else:
                key = generate_rsa_key()
                bin = key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption(),
                )
                pathlib.Path(key_path).write_bytes(bin)
            self._key_cache[name] = key
        return self._key_cache[name]

    def sign(self, name: str, cb: x509.CertificateBuilder) -> x509.Certificate:
        backend = cryptography.hazmat.backends.default_backend()
        cert = cb.issuer_name(self._root_ca.subject) \
            .sign(private_key=self._root_key, algorithm=hashes.SHA256(), backend=backend)
        cert_path = self.cert_path(name)
        pathlib.Path(cert_path).write_bytes(cert.public_bytes(serialization.Encoding.PEM))
        return cert


def generate_rsa_key() -> rsa.RSAPrivateKeyWithSerialization:
    backend = cryptography.hazmat.backends.default_backend()
    return rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=backend)


def generate_root_certificate(private_key: rsa.RSAPrivateKey) -> x509.Certificate:
    backend = cryptography.hazmat.backends.default_backend()
    subject = x509.Name(
        [x509.NameAttribute(
            x509.oid.NameOID.COMMON_NAME, 'root'
        )]
    )
    builder = x509.CertificateBuilder()
    return builder.subject_name(subject).issuer_name(subject) \
        .not_valid_before(datetime.datetime.utcnow()) \
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=1)) \
        .serial_number(1) \
        .public_key(private_key.public_key()) \
        .add_extension(extension=x509.BasicConstraints(ca=True, path_length=1), critical=True) \
        .add_extension(extension=x509.KeyUsage(digital_signature=False,
                                               content_commitment=False,
                                               key_encipherment=False,
                                               data_encipherment=False,
                                               key_agreement=False,
                                               key_cert_sign=True,
                                               crl_sign=True,
                                               encipher_only=False,
                                               decipher_only=False,
                                               ), critical=True) \
        .sign(private_key=private_key, algorithm=hashes.SHA256(), backend=backend)




# vim: set fileencoding=utf8 :
"""
.. module:: login
   :platform: Unix, Windows, MacOSX
   :synopsis: Login classes

.. moduleauthor:: Mikhail Denisenko <denisenkom@gmail.com>
"""
import socket
import logging

logger = logging.getLogger(__name__)


class SspiAuth(object):
    """ SSPI authentication

    :platform: Windows

    Required parameters are server_name and port or spn

    :keyword user_name: User name, if not provided current security context will be used
    :type user_name: str
    :keyword password: User password, if not provided current security context will be used
    :type password: str
    :keyword server_name: MSSQL server host name
    :type server_name: str
    :keyword port: MSSQL server port
    :type port: int
    :keyword spn: Service name
    :type spn: str
    """
    def __init__(self, user_name='', password='', server_name='', port=None, spn=None):
        from . import sspi
        # parse username/password informations
        if '\\' in user_name:
            domain, user_name = user_name.split('\\')
        else:
            domain = ''
        if domain and user_name:
            self._identity = sspi.make_winnt_identity(
                domain,
                user_name,
                password)
        else:
            self._identity = None
        # build SPN
        if spn:
            self._sname = spn
        else:
            primary_host_name, _, _ = socket.gethostbyname_ex(server_name)
            self._sname = 'MSSQLSvc/{0}:{1}'.format(primary_host_name, port)

        # using Negotiate system will use proper protocol (either NTLM or Kerberos)
        self._cred = sspi.SspiCredentials(
            package='Negotiate',
            use=sspi.SECPKG_CRED_OUTBOUND,
            identity=self._identity)

        self._flags = sspi.ISC_REQ_CONFIDENTIALITY | sspi.ISC_REQ_REPLAY_DETECT | sspi.ISC_REQ_CONNECTION
        self._ctx = None

    def create_packet(self):
        from . import sspi
        import ctypes
        buf = ctypes.create_string_buffer(4096)
        self._ctx, status, bufs = self._cred.create_context(
            flags=self._flags,
            byte_ordering='network',
            target_name=self._sname,
            output_buffers=[(sspi.SECBUFFER_TOKEN, buf)])
        if status == sspi.Status.SEC_I_COMPLETE_AND_CONTINUE:
            self._ctx.complete_auth_token(bufs)
        return bufs[0][1]

    def handle_next(self, packet):
        from . import sspi
        import ctypes
        buf = ctypes.create_string_buffer(4096)
        status, buffers = self._ctx.next(
            flags=self._flags,
            byte_ordering='network',
            target_name=self._sname,
            input_buffers=[(sspi.SECBUFFER_TOKEN, packet)],
            output_buffers=[(sspi.SECBUFFER_TOKEN, buf)])
        return buffers[0][1]

    def close(self):
        self._ctx.close()
        self._cred.close()


class NtlmAuth(object):
    """ NTLM authentication, uses Python implementation

    :param user_name: User name
    :type user_name: str
    :param password: User password
    :type password: str
    """
    def __init__(self, user_name, password):
        self._user_name = user_name
        if '\\' in user_name:
            domain, self._user = user_name.split('\\', 1)
            self._domain = domain.upper()
        else:
            self._domain = 'WORKSPACE'
            self._user = user_name
        self._password = password
        try:
            from ntlm_auth.ntlm import NegotiateFlags
        except ImportError:
            raise ImportError("To use NTLM authentication you need to install ntlm-auth module")
        self._nego_flags = NegotiateFlags.NTLMSSP_NEGOTIATE_128 | \
                           NegotiateFlags.NTLMSSP_NEGOTIATE_56 | \
                           NegotiateFlags.NTLMSSP_NEGOTIATE_UNICODE | \
                           NegotiateFlags.NTLMSSP_NEGOTIATE_VERSION | \
                           NegotiateFlags.NTLMSSP_REQUEST_TARGET | \
                           NegotiateFlags.NTLMSSP_NEGOTIATE_NTLM | \
                           NegotiateFlags.NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY | \
                           NegotiateFlags.NTLMSSP_NEGOTIATE_ALWAYS_SIGN
        self._ntlm_compat = 2
        self._workstation = socket.gethostname().upper()

    def create_packet(self):
        import ntlm_auth.ntlm
        return ntlm_auth.ntlm.NegotiateMessage(
            negotiate_flags=self._nego_flags,
            domain_name=self._domain,
            workstation=self._workstation,
        ).get_data()

    def handle_next(self, packet):
        import ntlm_auth.ntlm
        challenge = ntlm_auth.ntlm.ChallengeMessage(packet)
        return ntlm_auth.ntlm.AuthenticateMessage(
            user_name=self._user,
            password=self._password,
            domain_name=self._domain,
            workstation=self._workstation,
            challenge_message=challenge,
            ntlm_compatibility=self._ntlm_compat,
            server_certificate_hash=None,
        ).get_data()

    def close(self):
        pass


class KerberosAuth(object):
    def __init__(self, server_principal):
        try:
            import kerberos
        except ImportError:
            import winkerberos as kerberos
        self._kerberos = kerberos
        res, context = kerberos.authGSSClientInit(server_principal)
        if res < 0:
            raise RuntimeError('authGSSClientInit failed with code {}'.format(res))
        logger.info('Initialized GSS context')
        self._context = context

    def create_packet(self):
        import base64
        res = self._kerberos.authGSSClientStep(self._context, '')
        if res < 0:
            raise RuntimeError('authGSSClientStep failed with code {}'.format(res))
        data = self._kerberos.authGSSClientResponse(self._context)
        logger.info('created first client GSS packet %s', data)
        return base64.b64decode(data)

    def handle_next(self, packet):
        import base64
        res = self._kerberos.authGSSClientStep(self._context, base64.b64encode(packet).decode('ascii'))
        if res < 0:
            raise RuntimeError('authGSSClientStep failed with code {}'.format(res))
        if res == self._kerberos.AUTH_GSS_COMPLETE:
            logger.info('GSS authentication completed')
            return b''
        else:
            data = self._kerberos.authGSSClientResponse(self._context)
            logger.info('created client GSS packet %s', data)
            return base64.b64decode(data)

    def close(self):
        pass



import six
from ctypes import c_ulong, c_ushort, c_void_p, c_ulonglong, POINTER,\
    Structure, c_wchar_p, WINFUNCTYPE, windll, byref, cast


class Status(object):
    SEC_E_OK = 0
    SEC_I_CONTINUE_NEEDED = 0x00090312
    SEC_I_COMPLETE_AND_CONTINUE = 0x00090314
    SEC_I_INCOMPLETE_CREDENTIALS = 0x00090320
    SEC_E_INSUFFICIENT_MEMORY = 0x80090300 - 0x100000000
    SEC_E_INVALID_HANDLE = 0x80090301 - 0x100000000
    SEC_E_UNSUPPORTED_FUNCTION = 0x80090302 - 0x100000000
    SEC_E_INTERNAL_ERROR = 0x80090304 - 0x100000000
    SEC_E_SECPKG_NOT_FOUND = 0x80090305 - 0x100000000
    SEC_E_NOT_OWNER = 0x80090306 - 0x100000000
    SEC_E_INVALID_TOKEN = 0x80090308 - 0x100000000
    SEC_E_NO_IMPERSONATION = 0x8009030B - 0x100000000
    SEC_E_LOGON_DENIED = 0x8009030C - 0x100000000
    SEC_E_UNKNOWN_CREDENTIALS = 0x8009030D - 0x100000000
    SEC_E_NO_CREDENTIALS = 0x8009030E - 0x100000000
    SEC_E_OUT_OF_SEQUENCE = 0x80090310 - 0x100000000
    SEC_E_NO_AUTHENTICATING_AUTHORITY = 0x80090311 - 0x100000000
    SEC_E_BUFFER_TOO_SMALL = 0x80090321 - 0x100000000
    SEC_E_WRONG_PRINCIPAL = 0x80090322 - 0x100000000
    SEC_E_ALGORITHM_MISMATCH = 0x80090331 - 0x100000000

    @classmethod
    def getname(cls, value):
        for name in dir(cls):
            if name.startswith('SEC_E_') and getattr(cls, name) == value:
                return name
        return 'unknown value {0:x}'.format(0x100000000 + value)

#define SECBUFFER_EMPTY             0   // Undefined, replaced by provider
#define SECBUFFER_DATA              1   // Packet data
SECBUFFER_TOKEN = 2
#define SECBUFFER_PKG_PARAMS        3   // Package specific parameters
#define SECBUFFER_MISSING           4   // Missing Data indicator
#define SECBUFFER_EXTRA             5   // Extra data
#define SECBUFFER_STREAM_TRAILER    6   // Security Trailer
#define SECBUFFER_STREAM_HEADER     7   // Security Header
#define SECBUFFER_NEGOTIATION_INFO  8   // Hints from the negotiation pkg
#define SECBUFFER_PADDING           9   // non-data padding
#define SECBUFFER_STREAM            10  // whole encrypted message
#define SECBUFFER_MECHLIST          11
#define SECBUFFER_MECHLIST_SIGNATURE 12
#define SECBUFFER_TARGET            13  // obsolete
#define SECBUFFER_CHANNEL_BINDINGS  14
#define SECBUFFER_CHANGE_PASS_RESPONSE 15
#define SECBUFFER_TARGET_HOST       16
#define SECBUFFER_ALERT             17

SECPKG_CRED_INBOUND = 0x00000001
SECPKG_CRED_OUTBOUND = 0x00000002
SECPKG_CRED_BOTH = 0x00000003
SECPKG_CRED_DEFAULT = 0x00000004
SECPKG_CRED_RESERVED = 0xF0000000

SECBUFFER_VERSION = 0

#define ISC_REQ_DELEGATE                0x00000001
#define ISC_REQ_MUTUAL_AUTH             0x00000002
ISC_REQ_REPLAY_DETECT = 4
#define ISC_REQ_SEQUENCE_DETECT         0x00000008
ISC_REQ_CONFIDENTIALITY = 0x10
ISC_REQ_USE_SESSION_KEY = 0x00000020
ISC_REQ_PROMPT_FOR_CREDS = 0x00000040
ISC_REQ_USE_SUPPLIED_CREDS = 0x00000080
ISC_REQ_ALLOCATE_MEMORY = 0x00000100
ISC_REQ_USE_DCE_STYLE = 0x00000200
ISC_REQ_DATAGRAM = 0x00000400
ISC_REQ_CONNECTION = 0x00000800
#define ISC_REQ_CALL_LEVEL              0x00001000
#define ISC_REQ_FRAGMENT_SUPPLIED       0x00002000
#define ISC_REQ_EXTENDED_ERROR          0x00004000
#define ISC_REQ_STREAM                  0x00008000
#define ISC_REQ_INTEGRITY               0x00010000
#define ISC_REQ_IDENTIFY                0x00020000
#define ISC_REQ_NULL_SESSION            0x00040000
#define ISC_REQ_MANUAL_CRED_VALIDATION  0x00080000
#define ISC_REQ_RESERVED1               0x00100000
#define ISC_REQ_FRAGMENT_TO_FIT         0x00200000
#// This exists only in Windows Vista and greater
#define ISC_REQ_FORWARD_CREDENTIALS     0x00400000
#define ISC_REQ_NO_INTEGRITY            0x00800000 // honored only by SPNEGO
#define ISC_REQ_USE_HTTP_STYLE          0x01000000
#define ISC_REQ_UNVERIFIED_TARGET_NAME  0x20000000
#define ISC_REQ_CONFIDENTIALITY_ONLY    0x40000000 // honored by SPNEGO/Kerberos

SECURITY_NETWORK_DREP = 0
SECURITY_NATIVE_DREP = 0x10

SECPKG_CRED_ATTR_NAMES = 1

ULONG = c_ulong
USHORT = c_ushort
PULONG = POINTER(ULONG)
PVOID = c_void_p
TimeStamp = c_ulonglong
PTimeStamp = POINTER(c_ulonglong)
PLUID = POINTER(c_ulonglong)


class SecHandle(Structure):
    _fields_ = [
        ('lower', c_void_p),
        ('upper', c_void_p),
    ]
PSecHandle = POINTER(SecHandle)
CredHandle = SecHandle
PCredHandle = PSecHandle
PCtxtHandle = PSecHandle


class SecBuffer(Structure):
    _fields_ = [
        ('cbBuffer', ULONG),
        ('BufferType', ULONG),
        ('pvBuffer', PVOID),
    ]
PSecBuffer = POINTER(SecBuffer)


class SecBufferDesc(Structure):
    _fields_ = [
        ('ulVersion', ULONG),
        ('cBuffers', ULONG),
        ('pBuffers', PSecBuffer),
    ]
PSecBufferDesc = POINTER(SecBufferDesc)


class SEC_WINNT_AUTH_IDENTITY(Structure):
    _fields_ = [
        ('User', c_wchar_p),
        ('UserLength', c_ulong),
        ('Domain', c_wchar_p),
        ('DomainLength', c_ulong),
        ('Password', c_wchar_p),
        ('PasswordLength', c_ulong),
        ('Flags', c_ulong),
        ]


class SecPkgInfo(Structure):
    _fields_ = [
        ('fCapabilities', ULONG),
        ('wVersion', USHORT),
        ('wRPCID', USHORT),
        ('cbMaxToken', ULONG),
        ('Name', c_wchar_p),
        ('Comment', c_wchar_p),
    ]
PSecPkgInfo = POINTER(SecPkgInfo)


class SecPkgCredentials_Names(Structure):
    _fields_ = [('UserName', c_wchar_p)]


def ret_val(value):
    if value < 0:
        raise Exception('SSPI Error {0}'.format(Status.getname(value)))
    return value


ENUMERATE_SECURITY_PACKAGES_FN = WINFUNCTYPE(
    ret_val,
    POINTER(c_ulong),
    POINTER(POINTER(SecPkgInfo)))

ACQUIRE_CREDENTIALS_HANDLE_FN = WINFUNCTYPE(
    ret_val,
    c_wchar_p,    # principal
    c_wchar_p,    # package
    ULONG,        # fCredentialUse
    PLUID,        # pvLogonID
    PVOID,        # pAuthData
    PVOID,        # pGetKeyFn
    PVOID,        # pvGetKeyArgument
    PCredHandle,  # phCredential
    PTimeStamp    # ptsExpiry
    )
FREE_CREDENTIALS_HANDLE_FN = WINFUNCTYPE(ret_val, POINTER(SecHandle))
INITIALIZE_SECURITY_CONTEXT_FN = WINFUNCTYPE(
    ret_val,
    PCredHandle,
    PCtxtHandle,     # phContext,
    c_wchar_p,       # pszTargetName,
    ULONG,           # fContextReq,
    ULONG,           # Reserved1,
    ULONG,           # TargetDataRep,
    PSecBufferDesc,  # pInput,
    ULONG,           # Reserved2,
    PCtxtHandle,     # phNewContext,
    PSecBufferDesc,  # pOutput,
    PULONG,          # pfContextAttr,
    PTimeStamp,      # ptsExpiry
    )
COMPLETE_AUTH_TOKEN_FN = WINFUNCTYPE(
    ret_val,
    PCtxtHandle,     # phContext
    PSecBufferDesc,  # pToken
    )

FREE_CONTEXT_BUFFER_FN = WINFUNCTYPE(ret_val, PVOID)

QUERY_CREDENTIAL_ATTRIBUTES_FN = WINFUNCTYPE(
    ret_val,
    PCredHandle,    # cred
    ULONG,          # attribute
    PVOID,          # out buffer
    )
ACCEPT_SECURITY_CONTEXT_FN = PVOID
DELETE_SECURITY_CONTEXT_FN = WINFUNCTYPE(ret_val, PCtxtHandle)
APPLY_CONTROL_TOKEN_FN = PVOID
QUERY_CONTEXT_ATTRIBUTES_FN = PVOID
IMPERSONATE_SECURITY_CONTEXT_FN = PVOID
REVERT_SECURITY_CONTEXT_FN = PVOID
MAKE_SIGNATURE_FN = PVOID
VERIFY_SIGNATURE_FN = PVOID
QUERY_SECURITY_PACKAGE_INFO_FN = WINFUNCTYPE(
    ret_val,
    c_wchar_p,  # package name
    POINTER(PSecPkgInfo),
    )
EXPORT_SECURITY_CONTEXT_FN = PVOID
IMPORT_SECURITY_CONTEXT_FN = PVOID
ADD_CREDENTIALS_FN = PVOID
QUERY_SECURITY_CONTEXT_TOKEN_FN = PVOID
ENCRYPT_MESSAGE_FN = PVOID
DECRYPT_MESSAGE_FN = PVOID
SET_CONTEXT_ATTRIBUTES_FN = PVOID


class SECURITY_FUNCTION_TABLE(Structure):
    _fields_ = [
        ('dwVersion', c_ulong),
        ('EnumerateSecurityPackages', ENUMERATE_SECURITY_PACKAGES_FN),
        ('QueryCredentialsAttributes', QUERY_CREDENTIAL_ATTRIBUTES_FN),
        ('AcquireCredentialsHandle', ACQUIRE_CREDENTIALS_HANDLE_FN),
        ('FreeCredentialsHandle', FREE_CREDENTIALS_HANDLE_FN),
        ('Reserved2', c_void_p),
        ('InitializeSecurityContext', INITIALIZE_SECURITY_CONTEXT_FN),
        ('AcceptSecurityContext', ACCEPT_SECURITY_CONTEXT_FN),
        ('CompleteAuthToken', COMPLETE_AUTH_TOKEN_FN),
        ('DeleteSecurityContext', DELETE_SECURITY_CONTEXT_FN),
        ('ApplyControlToken', APPLY_CONTROL_TOKEN_FN),
        ('QueryContextAttributes', QUERY_CONTEXT_ATTRIBUTES_FN),
        ('ImpersonateSecurityContext', IMPERSONATE_SECURITY_CONTEXT_FN),
        ('RevertSecurityContext', REVERT_SECURITY_CONTEXT_FN),
        ('MakeSignature', MAKE_SIGNATURE_FN),
        ('VerifySignature', VERIFY_SIGNATURE_FN),
        ('FreeContextBuffer', FREE_CONTEXT_BUFFER_FN),
        ('QuerySecurityPackageInfo', QUERY_SECURITY_PACKAGE_INFO_FN),
        ('Reserved3', c_void_p),
        ('Reserved4', c_void_p),
        ('ExportSecurityContext', EXPORT_SECURITY_CONTEXT_FN),
        ('ImportSecurityContext', IMPORT_SECURITY_CONTEXT_FN),
        ('AddCredentials', ADD_CREDENTIALS_FN),
        ('Reserved8', c_void_p),
        ('QuerySecurityContextToken', QUERY_SECURITY_CONTEXT_TOKEN_FN),
        ('EncryptMessage', ENCRYPT_MESSAGE_FN),
        ('DecryptMessage', DECRYPT_MESSAGE_FN),
        ('SetContextAttributes', SET_CONTEXT_ATTRIBUTES_FN),
        ]

_PInitSecurityInterface = WINFUNCTYPE(POINTER(SECURITY_FUNCTION_TABLE))
InitSecurityInterface = _PInitSecurityInterface(('InitSecurityInterfaceW', windll.secur32))

sec_fn = InitSecurityInterface()
if not sec_fn:
    raise Exception('InitSecurityInterface failed')
sec_fn = sec_fn.contents


class _SecContext(object):
    def close(self):
        if self._handle.lower and self._handle.upper:
            sec_fn.DeleteSecurityContext(self._handle)
            self._handle.lower = self._handle.upper = 0

    def __del__(self):
        self.close()

    def complete_auth_token(self, bufs):
        sec_fn.CompleteAuthToken(
            byref(self._handle),
            byref(_make_buffers_desc(bufs)))

    def next(self,
             flags,
             target_name=None,
             byte_ordering='network',
             input_buffers=None,
             output_buffers=None):
        input_buffers_desc = _make_buffers_desc(input_buffers) if input_buffers else None
        output_buffers_desc = _make_buffers_desc(output_buffers) if output_buffers else None
        status = sec_fn.InitializeSecurityContext(
            byref(self._cred._handle),
            byref(self._handle),
            target_name,
            flags,
            0,
            SECURITY_NETWORK_DREP if byte_ordering == 'network' else SECURITY_NATIVE_DREP,
            byref(input_buffers_desc) if input_buffers_desc else None,
            0,
            byref(self._handle),
            byref(output_buffers_desc) if input_buffers_desc else None,
            byref(self._attrs),
            byref(self._ts))
        result_buffers = []
        for i, (type, buf) in enumerate(output_buffers):
            buf = buf[:output_buffers_desc.pBuffers[i].cbBuffer]
            result_buffers.append((type, buf))
        return status, result_buffers


class SspiCredentials(object):
    def __init__(self, package, use, identity=None):
        self._handle = SecHandle()
        self._ts = TimeStamp()
        sec_fn.AcquireCredentialsHandle(
            None, package, use,
            None, byref(identity) if identity and identity.Domain else None,
            None, None, byref(self._handle), byref(self._ts))

    def close(self):
        if self._handle.lower or self._handle.upper:
            sec_fn.FreeCredentialsHandle(byref(self._handle))
            self._handle.lower = 0
            self._handle.upper = 0

    def __del__(self):
        self.close()

    def query_user_name(self):
        names = SecPkgCredentials_Names()
        try:
            sec_fn.QueryCredentialsAttributes(
                byref(self._handle),
                SECPKG_CRED_ATTR_NAMES,
                byref(names))
            user_name = six.text_type(names.UserName)
        finally:
            p = c_wchar_p.from_buffer(names, SecPkgCredentials_Names.UserName.offset)
            sec_fn.FreeContextBuffer(p)
        return user_name

    def create_context(
            self,
            flags,
            target_name=None,
            byte_ordering='network',
            input_buffers=None,
            output_buffers=None):
        ctx = _SecContext()
        ctx._cred = self
        ctx._handle = SecHandle()
        ctx._ts = TimeStamp()
        ctx._attrs = ULONG()
        input_buffers_desc = _make_buffers_desc(input_buffers) if input_buffers else None
        output_buffers_desc = _make_buffers_desc(output_buffers) if output_buffers else None
        status = sec_fn.InitializeSecurityContext(
            byref(self._handle),
            None,
            target_name,
            flags,
            0,
            SECURITY_NETWORK_DREP if byte_ordering == 'network' else SECURITY_NATIVE_DREP,
            byref(input_buffers_desc) if input_buffers_desc else None,
            0,
            byref(ctx._handle),
            byref(output_buffers_desc) if output_buffers_desc else None,
            byref(ctx._attrs),
            byref(ctx._ts))
        result_buffers = []
        for i, (type, buf) in enumerate(output_buffers):
            buf = buf[:output_buffers_desc.pBuffers[i].cbBuffer]
            result_buffers.append((type, buf))
        return ctx, status, result_buffers


def _make_buffers_desc(buffers):
    desc = SecBufferDesc()
    desc.ulVersion = SECBUFFER_VERSION
    bufs_array = (SecBuffer * len(buffers))()
    for i, (type, buf) in enumerate(buffers):
        bufs_array[i].BufferType = type
        bufs_array[i].cbBuffer = len(buf)
        bufs_array[i].pvBuffer = cast(buf, PVOID)
    desc.pBuffers = bufs_array
    desc.cBuffers = len(buffers)
    return desc


def make_winnt_identity(domain, user_name, password):
    identity = SEC_WINNT_AUTH_IDENTITY()
    identity.Flags = 2  # SEC_WINNT_AUTH_IDENTITY_UNICODE
    identity.Password = password
    identity.PasswordLength = len(password)
    identity.Domain = domain
    identity.DomainLength = len(domain)
    identity.User = user_name
    identity.UserLength = len(user_name)
    return identity

#class SspiSecBuffer(object):
#    def __init__(self, type, buflen=4096):
#        self._buf = create_string_buffer(int(buflen))
#        self._desc = SecBuffer()
#        self._desc.cbBuffer = buflen
#        self._desc.BufferType = type
#        self._desc.pvBuffer = cast(self._buf, PVOID)
#
#class SspiSecBuffers(object):
#    def __init__(self):
#        self._desc = SecBufferDesc()
#        self._desc.ulVersion = SECBUFFER_VERSION
#        self._descrs = (SecBuffer * 8)()
#        self._desc.pBuffers = self._descrs
#
#    def append(self, buf):
#        if len(self._descrs) <= self._desc.cBuffers:
#            newdescrs = (SecBuffer * (len(self._descrs) * 2))(*self._descrs)
#            self._descrs = newdescrs
#            self._desc.pBuffers = newdescrs
#        self._descrs[self._desc.cBuffers] = buf._desc
#        self._desc.cBuffers += 1


def enum_security_packages():
    num = ULONG()
    infos = POINTER(SecPkgInfo)()
    status = sec_fn.EnumerateSecurityPackages(byref(num), byref(infos))
    try:
        return [{'caps': infos[i].fCapabilities,
                 'version': infos[i].wVersion,
                 'rpcid': infos[i].wRPCID,
                 'max_token': infos[i].cbMaxToken,
                 'name': infos[i].Name,
                 'comment': infos[i].Comment,
                 } for i in range(num.value)]
    finally:
        sec_fn.FreeContextBuffer(infos)



import logging

import pytest

import pytds.tds
import settings


logger = logging.getLogger(__name__)
LIVE_TEST = getattr(settings, 'LIVE_TEST', True)
pytds.tds.logging_enabled = True


@pytest.fixture(scope='module')
def db_connection():
    if not LIVE_TEST:
        pytest.skip('LIVE_TEST is not set')
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs['database'] = 'master'
    return pytds.connect(*settings.CONNECT_ARGS, **kwargs)


@pytest.fixture
def cursor(db_connection):
    with db_connection.cursor() as cursor:
        yield cursor
    db_connection.rollback()


@pytest.fixture
def separate_db_connection():
    if not LIVE_TEST:
        pytest.skip('LIVE_TEST is not set')
    kwargs = settings.CONNECT_KWARGS.copy()
    kwargs['database'] = 'master'
    conn = pytds.connect(*settings.CONNECT_ARGS, **kwargs)
    yield conn
    conn.close()


@pytest.fixture(scope='module')
def collation_set(db_connection):
    with db_connection.cursor() as cursor:
        cursor.execute("SELECT Name, Description, COLLATIONPROPERTY(Name, 'LCID') FROM ::fn_helpcollations()")
        collations_list = cursor.fetchall()
    return set(coll_name for coll_name, _, _ in collations_list)



import sys
import unittest


if sys.version_info.major < 3:
    # patch unittest.TestCase for Python 2.7
    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp


class MockSock(object):
    def __init__(self, input_packets=()):
        self.set_input(input_packets)
        self._out_packets = []
        self._closed = False

    def recv(self, size):
        if not self.is_open():
            raise Exception('Connection closed')
        if self._curr_packet >= len(self._packets):
            return b''
        if self._packet_pos >= len(self._packets[self._curr_packet]):
            self._curr_packet += 1
            self._packet_pos = 0
        if self._curr_packet >= len(self._packets):
            return b''
        res = self._packets[self._curr_packet][self._packet_pos:self._packet_pos+size]
        self._packet_pos += len(res)
        return res

    def recv_into(self, buffer, size=0):
        if not self.is_open():
            raise Exception('Connection closed')
        if size == 0:
            size = len(buffer)
        res = self.recv(size)
        buffer[0:len(res)] = res
        return len(res)

    def send(self, buf, flags=0):
        if not self.is_open():
            raise Exception('Connection closed')
        self._out_packets.append(buf)
        return len(buf)

    def sendall(self, buf, flags=0):
        if not self.is_open():
            raise Exception('Connection closed')
        self._out_packets.append(buf)

    def setsockopt(self, *args):
        pass

    def close(self):
        self._closed = True

    def is_open(self):
        return not self._closed

    def consume_output(self):
        """
        Retrieve data from output queue and then clear output queue
        @return: bytes
        """
        res = self._out_packets
        self._out_packets = []
        return b''.join(res)

    def set_input(self, packets):
        """
        Resets input queue
        @param packets: List of input packets
        """
        self._packets = packets
        self._curr_packet = 0
        self._packet_pos = 0



import itertools
import datetime
import decimal
import struct
import re
import uuid
import six
import functools
from io import StringIO, BytesIO

from pytds.tds_base import read_chunks
from . import tds_base
from .collate import ucs2_codec, raw_collation
from . import tz


_flt4_struct = struct.Struct('f')
_flt8_struct = struct.Struct('d')
_utc = tz.utc


def _applytz(dt, tzinfo):
    if not tzinfo:
        return dt
    dt = dt.replace(tzinfo=tzinfo)
    return dt


def _decode_num(buf):
    """ Decodes little-endian integer from buffer

    Buffer can be of any size
    """
    return functools.reduce(lambda acc, val: acc * 256 + tds_base.my_ord(val), reversed(buf), 0)


class PlpReader(object):
    """ Partially length prefixed reader

    Spec: http://msdn.microsoft.com/en-us/library/dd340469.aspx
    """
    def __init__(self, r):
        """
        :param r: An instance of :class:`_TdsReader`
        """
        self._rdr = r
        size = r.get_uint8()
        self._size = size

    def is_null(self):
        """
        :return: True if stored value is NULL
        """
        return self._size == tds_base.PLP_NULL

    def is_unknown_len(self):
        """
        :return: True if total size is unknown upfront
        """
        return self._size == tds_base.PLP_UNKNOWN

    def size(self):
        """
        :return: Total size in bytes if is_uknown_len and is_null are both False
        """
        return self._size

    def chunks(self):
        """ Generates chunks from stream, each chunk is an instace of bytes.
        """
        if self.is_null():
            return
        total = 0
        while True:
            chunk_len = self._rdr.get_uint()
            if chunk_len == 0:
                if not self.is_unknown_len() and total != self._size:
                    msg = "PLP actual length (%d) doesn't match reported length (%d)" % (total, self._size)
                    self._rdr.session.bad_stream(msg)

                return

            total += chunk_len
            left = chunk_len
            while left:
                buf = self._rdr.recv(left)
                yield buf
                left -= len(buf)


class _StreamChunkedHandler(object):
    def __init__(self, stream):
        self.stream = stream

    def add_chunk(self, val):
        self.stream.write(val)

    def end(self):
        return self.stream


class _DefaultChunkedHandler(object):
    def __init__(self, stream):
        self.stream = stream

    def add_chunk(self, val):
        self.stream.write(val)

    def end(self):
        value = self.stream.getvalue()
        self.stream.seek(0)
        self.stream.truncate()
        return value

    def __eq__(self, other):
        return self.stream.getvalue() == other.stream.getvalue()

    def __ne__(self, other):
        return not self.__eq__(other)


class SqlTypeMetaclass(tds_base.CommonEqualityMixin):
    def __repr__(self):
        return '<sqltype:{}>'.format(self.get_declaration())

    def get_declaration(self):
        raise NotImplementedError()


class ImageType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'IMAGE'


class BinaryType(SqlTypeMetaclass):
    def __init__(self, size=30):
        self._size = size

    @property
    def size(self):
        return self._size

    def get_declaration(self):
        return 'BINARY({})'.format(self._size)


class VarBinaryType(SqlTypeMetaclass):
    def __init__(self, size=30):
        self._size = size

    @property
    def size(self):
        return self._size

    def get_declaration(self):
        return 'VARBINARY({})'.format(self._size)


class VarBinaryMaxType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'VARBINARY(MAX)'


class CharType(SqlTypeMetaclass):
    def __init__(self, size=30):
        self._size = size

    @property
    def size(self):
        return self._size

    def get_declaration(self):
        return 'CHAR({})'.format(self._size)


class VarCharType(SqlTypeMetaclass):
    def __init__(self, size=30):
        self._size = size

    @property
    def size(self):
        return self._size

    def get_declaration(self):
        return 'VARCHAR({})'.format(self._size)


class VarCharMaxType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'VARCHAR(MAX)'


class NCharType(SqlTypeMetaclass):
    def __init__(self, size=30):
        self._size = size

    @property
    def size(self):
        return self._size

    def get_declaration(self):
        return 'NCHAR({})'.format(self._size)


class NVarCharType(SqlTypeMetaclass):
    def __init__(self, size=30):
        self._size = size

    @property
    def size(self):
        return self._size

    def get_declaration(self):
        return 'NVARCHAR({})'.format(self._size)


class NVarCharMaxType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'NVARCHAR(MAX)'


class TextType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'TEXT'


class NTextType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'NTEXT'


class XmlType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'XML'


class SmallMoneyType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'SMALLMONEY'


class MoneyType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'MONEY'


class DecimalType(SqlTypeMetaclass):
    def __init__(self, precision=18, scale=0):
        self._precision = precision
        self._scale = scale

    @classmethod
    def from_value(cls, value):
        if not (-10 ** 38 + 1 <= value <= 10 ** 38 - 1):
            raise tds_base.DataError('Decimal value is out of range')
        with decimal.localcontext() as context:
            context.prec = 38
            value = value.normalize()
            _, digits, exp = value.as_tuple()
            if exp > 0:
                scale = 0
                prec = len(digits) + exp
            else:
                scale = -exp
                prec = max(len(digits), scale)
            return cls(precision=prec, scale=scale)

    @property
    def precision(self):
        return self._precision

    @property
    def scale(self):
        return self._scale

    def get_declaration(self):
        return 'DECIMAL({}, {})'.format(self._precision, self._scale)


class UniqueIdentifierType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'UNIQUEIDENTIFIER'


class VariantType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'SQL_VARIANT'


class SqlValueMetaclass(tds_base.CommonEqualityMixin):
    pass


class BaseTypeSerializer(tds_base.CommonEqualityMixin):
    """ Base type for TDS data types.

    All TDS types should derive from it.
    In addition actual types should provide the following:

    - type - class variable storing type identifier
    """
    type = 0

    def __init__(self, precision=None, scale=None, size=None):
        self._precision = precision
        self._scale = scale
        self._size = size

    @property
    def precision(self):
        return self._precision

    @property
    def scale(self):
        return self._scale

    @property
    def size(self):
        return self._size

    def get_typeid(self):
        """ Returns type identifier of type. """
        return self.type

    @classmethod
    def from_stream(cls, r):
        """ Class method that reads and returns a type instance.

        :param r: An instance of :class:`_TdsReader` to read type from.

        Should be implemented in actual types.
        """
        raise NotImplementedError

    def write_info(self, w):
        """ Writes type info into w stream.

        :param w: An instance of :class:`_TdsWriter` to write into.

        Should be symmetrical to from_stream method.
        Should be implemented in actual types.
        """
        raise NotImplementedError

    def write(self, w, value):
        """ Writes type's value into stream

        :param w: An instance of :class:`_TdsWriter` to write into.
        :param value: A value to be stored, should be compatible with the type

        Should be implemented in actual types.
        """
        raise NotImplementedError

    def read(self, r):
        """ Reads value from the stream.

        :param r: An instance of :class:`_TdsReader` to read value from.
        :return: A read value.

        Should be implemented in actual types.
        """
        raise NotImplementedError

    def set_chunk_handler(self, chunk_handler):
        raise ValueError("Column type does not support chunk handler")


class BasePrimitiveTypeSerializer(BaseTypeSerializer):
    """ Base type for primitive TDS data types.

    Primitive type is a fixed size type with no type arguments.
    All primitive TDS types should derive from it.
    In addition actual types should provide the following:

    - type - class variable storing type identifier
    - declaration - class variable storing name of sql type
    - isntance - class variable storing instance of class
    """

    def write(self, w, value):
        raise NotImplementedError

    def read(self, r):
        raise NotImplementedError

    instance = None

    @classmethod
    def from_stream(cls, r):
        return cls.instance

    def write_info(self, w):
        pass


class BaseTypeSerializerN(BaseTypeSerializer):
    """ Base type for nullable TDS data types.

    All nullable TDS types should derive from it.
    In addition actual types should provide the following:

    - type - class variable storing type identifier
    - subtypes - class variable storing dict {subtype_size: subtype_instance}
    """
    subtypes = {}

    def __init__(self, size):
        super(BaseTypeSerializerN, self).__init__(size=size)
        assert size in self.subtypes
        self._current_subtype = self.subtypes[size]

    def get_typeid(self):
        return self._current_subtype.get_typeid()

    @classmethod
    def from_stream(cls, r):
        size = r.get_byte()
        if size not in cls.subtypes:
            raise tds_base.InterfaceError('Invalid %s size' % cls.type, size)
        return cls(size)

    def write_info(self, w):
        w.put_byte(self.size)

    def read(self, r):
        size = r.get_byte()
        if size == 0:
            return None
        if size not in self.subtypes:
            raise r.session.bad_stream('Invalid %s size' % self.type, size)
        return self.subtypes[size].read(r)

    def write(self, w, val):
        if val is None:
            w.put_byte(0)
            return
        w.put_byte(self.size)
        self._current_subtype.write(w, val)


class BitType(SqlTypeMetaclass):
    type = tds_base.SYBBITN

    def get_declaration(self):
        return 'BIT'


class TinyIntType(SqlTypeMetaclass):
    type = tds_base.SYBINTN
    size = 1

    def get_declaration(self):
        return 'TINYINT'


class SmallIntType(SqlTypeMetaclass):
    type = tds_base.SYBINTN
    size = 2

    def get_declaration(self):
        return 'SMALLINT'


class IntType(SqlTypeMetaclass):
    type = tds_base.SYBINTN
    size = 4

    def get_declaration(self):
        return 'INT'


class BigIntType(SqlTypeMetaclass):
    type = tds_base.SYBINTN
    size = 8

    def get_declaration(self):
        return 'BIGINT'


class RealType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'REAL'


class FloatType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'FLOAT'


class BitSerializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBBIT
    declaration = 'BIT'

    def write(self, w, value):
        w.put_byte(1 if value else 0)

    def read(self, r):
        return bool(r.get_byte())

BitSerializer.instance = BitSerializer()


class BitNSerializer(BaseTypeSerializerN):
    type = tds_base.SYBBITN
    subtypes = {1: BitSerializer.instance}

    def __init__(self, typ):
        super(BitNSerializer, self).__init__(size=1)
        self._typ = typ

    def __repr__(self):
        return 'BitNSerializer({})'.format(self._typ)


BitNSerializer.instance = BitNSerializer(BitType())


class TinyIntSerializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBINT1
    declaration = 'TINYINT'

    def write(self, w, val):
        w.put_byte(val)

    def read(self, r):
        return r.get_byte()

TinyIntSerializer.instance = TinyIntSerializer()


class SmallIntSerializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBINT2
    declaration = 'SMALLINT'

    def write(self, w, val):
        w.put_smallint(val)

    def read(self, r):
        return r.get_smallint()

SmallIntSerializer.instance = SmallIntSerializer()


class IntSerializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBINT4
    declaration = 'INT'

    def write(self, w, val):
        w.put_int(val)

    def read(self, r):
        return r.get_int()

IntSerializer.instance = IntSerializer()


class BigIntSerializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBINT8
    declaration = 'BIGINT'

    def write(self, w, val):
        w.put_int8(val)

    def read(self, r):
        return r.get_int8()

BigIntSerializer.instance = BigIntSerializer()


class IntNSerializer(BaseTypeSerializerN):
    type = tds_base.SYBINTN

    subtypes = {
        1: TinyIntSerializer.instance,
        2: SmallIntSerializer.instance,
        4: IntSerializer.instance,
        8: BigIntSerializer.instance,
    }

    type_by_size = {
        1: TinyIntType(),
        2: SmallIntType(),
        4: IntType(),
        8: BigIntType(),
    }

    def __init__(self, typ):
        super(IntNSerializer, self).__init__(size=typ.size)
        self._typ = typ

    @classmethod
    def from_stream(cls, r):
        size = r.get_byte()
        if size not in cls.subtypes:
            raise tds_base.InterfaceError('Invalid %s size' % cls.type, size)
        return cls(cls.type_by_size[size])

    def __repr__(self):
        return 'IntN({})'.format(self.size)


class RealSerializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBREAL
    declaration = 'REAL'

    def write(self, w, val):
        w.pack(_flt4_struct, val)

    def read(self, r):
        return r.unpack(_flt4_struct)[0]

RealSerializer.instance = RealSerializer()


class FloatSerializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBFLT8
    declaration = 'FLOAT'

    def write(self, w, val):
        w.pack(_flt8_struct, val)

    def read(self, r):
        return r.unpack(_flt8_struct)[0]

FloatSerializer.instance = FloatSerializer()


class FloatNSerializer(BaseTypeSerializerN):
    type = tds_base.SYBFLTN

    subtypes = {
        4: RealSerializer.instance,
        8: FloatSerializer.instance,
    }


class VarChar(SqlValueMetaclass):
    def __init__(self, val, collation=raw_collation):
        self._val = val
        self._collation = collation

    @property
    def collation(self):
        return self._collation

    @property
    def val(self):
        return self._val

    def __str__(self):
        return self._val


class VarChar70Serializer(BaseTypeSerializer):
    type = tds_base.XSYBVARCHAR

    def __init__(self, size, collation=raw_collation, codec=None):
        super(VarChar70Serializer, self).__init__(size=size)
        self._collation = collation
        if codec:
            self._codec = codec
        else:
            self._codec = collation.get_codec()

    @classmethod
    def from_stream(cls, r):
        size = r.get_smallint()
        return cls(size, codec=r.session.conn.server_codec)

    def write_info(self, w):
        w.put_smallint(self.size)

    def write(self, w, val):
        if val is None:
            w.put_smallint(-1)
        else:
            if w._tds._tds._login.bytes_to_unicode:
                val = tds_base.force_unicode(val)
            if isinstance(val, six.text_type):
                val, _ = self._codec.encode(val)
            w.put_smallint(len(val))
            w.write(val)

    def read(self, r):
        size = r.get_smallint()
        if size < 0:
            return None
        if r._session._tds._login.bytes_to_unicode:
            return r.read_str(size, self._codec)
        else:
            return tds_base.readall(r, size)


class VarChar71Serializer(VarChar70Serializer):
    @classmethod
    def from_stream(cls, r):
        size = r.get_smallint()
        collation = r.get_collation()
        return cls(size, collation)

    def write_info(self, w):
        super(VarChar71Serializer, self).write_info(w)
        w.put_collation(self._collation)


class VarChar72Serializer(VarChar71Serializer):
    @classmethod
    def from_stream(cls, r):
        size = r.get_usmallint()
        collation = r.get_collation()
        if size == 0xffff:
            return VarCharMaxSerializer(collation)
        return cls(size, collation)


class VarCharMaxSerializer(VarChar72Serializer):
    def __init__(self, collation=raw_collation):
        super(VarChar72Serializer, self).__init__(0, collation)
        self._chunk_handler = None

    def write_info(self, w):
        w.put_usmallint(tds_base.PLP_MARKER)
        w.put_collation(self._collation)

    def write(self, w, val):
        if val is None:
            w.put_uint8(tds_base.PLP_NULL)
        else:
            if w._tds._tds._login.bytes_to_unicode:
                val = tds_base.force_unicode(val)
            if isinstance(val, six.text_type):
                val, _ = self._codec.encode(val)

            # Putting the actual length here causes an error when bulk inserting:
            #
            # While reading current row from host, a premature end-of-message
            # was encountered--an incoming data stream was interrupted when
            # the server expected to see more data. The host program may have
            # terminated. Ensure that you are using a supported client
            # application programming interface (API).
            #
            # See https://github.com/tediousjs/tedious/issues/197
            # It is not known why this happens, but Microsoft's bcp tool
            # uses PLP_UNKNOWN for varchar(max) as well.
            w.put_uint8(tds_base.PLP_UNKNOWN)
            if len(val) > 0:
                w.put_uint(len(val))
                w.write(val)
            w.put_uint(0)

    def read(self, r):
        login = r._session._tds._login
        r = PlpReader(r)
        if r.is_null():
            return None
        if self._chunk_handler is None:
            if login.bytes_to_unicode:
                self._chunk_handler = _DefaultChunkedHandler(StringIO())
            else:
                self._chunk_handler = _DefaultChunkedHandler(BytesIO())
        if login.bytes_to_unicode:
            for chunk in tds_base.iterdecode(r.chunks(), self._codec):
                self._chunk_handler.add_chunk(chunk)
        else:
            for chunk in r.chunks():
                self._chunk_handler.add_chunk(chunk)
        return self._chunk_handler.end()

    def set_chunk_handler(self, chunk_handler):
        self._chunk_handler = chunk_handler


class NVarChar70Serializer(BaseTypeSerializer):
    type = tds_base.XSYBNVARCHAR

    def __init__(self, size, collation=raw_collation):
        super(NVarChar70Serializer, self).__init__(size=size)
        self._collation = collation

    @classmethod
    def from_stream(cls, r):
        size = r.get_usmallint()
        return cls(size / 2)

    def write_info(self, w):
        w.put_usmallint(self.size * 2)

    def write(self, w, val):
        if val is None:
            w.put_usmallint(0xffff)
        else:
            if isinstance(val, bytes):
                val = tds_base.force_unicode(val)
            buf, _ = ucs2_codec.encode(val)
            l = len(buf)
            w.put_usmallint(l)
            w.write(buf)

    def read(self, r):
        size = r.get_usmallint()
        if size == 0xffff:
            return None
        return r.read_str(size, ucs2_codec)


class NVarChar71Serializer(NVarChar70Serializer):
    @classmethod
    def from_stream(cls, r):
        size = r.get_usmallint()
        collation = r.get_collation()
        return cls(size / 2, collation)

    def write_info(self, w):
        super(NVarChar71Serializer, self).write_info(w)
        w.put_collation(self._collation)


class NVarChar72Serializer(NVarChar71Serializer):
    @classmethod
    def from_stream(cls, r):
        size = r.get_usmallint()
        collation = r.get_collation()
        if size == 0xffff:
            return NVarCharMaxSerializer(collation=collation)
        return cls(size / 2, collation=collation)


class NVarCharMaxSerializer(NVarChar72Serializer):
    def __init__(self, collation=raw_collation):
        super(NVarCharMaxSerializer, self).__init__(size=-1, collation=collation)
        self._chunk_handler = _DefaultChunkedHandler(StringIO())

    def __repr__(self):
        return 'NVarCharMax(s={},c={})'.format(self.size, repr(self._collation))

    def get_typeid(self):
        return tds_base.SYBNTEXT

    def write_info(self, w):
        w.put_usmallint(tds_base.PLP_MARKER)
        w.put_collation(self._collation)

    def write(self, w, val):
        if val is None:
            w.put_uint8(tds_base.PLP_NULL)
        else:
            if isinstance(val, bytes):
                val = tds_base.force_unicode(val)
            val, _ = ucs2_codec.encode(val)

            # Putting the actual length here causes an error when bulk inserting:
            #
            # While reading current row from host, a premature end-of-message
            # was encountered--an incoming data stream was interrupted when
            # the server expected to see more data. The host program may have
            # terminated. Ensure that you are using a supported client
            # application programming interface (API).
            #
            # See https://github.com/tediousjs/tedious/issues/197
            # It is not known why this happens, but Microsoft's bcp tool
            # uses PLP_UNKNOWN for nvarchar(max) as well.
            w.put_uint8(tds_base.PLP_UNKNOWN)
            if len(val) > 0:
                w.put_uint(len(val))
                w.write(val)
            w.put_uint(0)

    def read(self, r):
        r = PlpReader(r)
        if r.is_null():
            return None
        for chunk in tds_base.iterdecode(r.chunks(), ucs2_codec):
            self._chunk_handler.add_chunk(chunk)
        return self._chunk_handler.end()

    def set_chunk_handler(self, chunk_handler):
        self._chunk_handler = chunk_handler


class XmlSerializer(NVarCharMaxSerializer):
    type = tds_base.SYBMSXML
    declaration = 'XML'

    def __init__(self, schema=None):
        super(XmlSerializer, self).__init__(0)
        self._schema = schema or {}

    def __repr__(self):
        return 'XmlSerializer(schema={})'.format(repr(self._schema))

    def get_typeid(self):
        return self.type

    @classmethod
    def from_stream(cls, r):
        has_schema = r.get_byte()
        schema = {}
        if has_schema:
            schema['dbname'] = r.read_ucs2(r.get_byte())
            schema['owner'] = r.read_ucs2(r.get_byte())
            schema['collection'] = r.read_ucs2(r.get_smallint())
        return cls(schema)

    def write_info(self, w):
        if self._schema:
            w.put_byte(1)
            w.put_byte(len(self._schema['dbname']))
            w.write_ucs2(self._schema['dbname'])
            w.put_byte(len(self._schema['owner']))
            w.write_ucs2(self._schema['owner'])
            w.put_usmallint(len(self._schema['collection']))
            w.write_ucs2(self._schema['collection'])
        else:
            w.put_byte(0)


class Text70Serializer(BaseTypeSerializer):
    type = tds_base.SYBTEXT
    declaration = 'TEXT'

    def __init__(self, size=0, table_name='', collation=raw_collation, codec=None):
        super(Text70Serializer, self).__init__(size=size)
        self._table_name = table_name
        self._collation = collation
        if codec:
            self._codec = codec
        else:
            self._codec = collation.get_codec()
        self._chunk_handler = None

    def __repr__(self):
        return 'Text70(size={},table_name={},codec={})'.format(self.size, self._table_name, self._codec)

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        table_name = r.read_ucs2(r.get_smallint())
        return cls(size, table_name, codec=r.session.conn.server_codec)

    def write_info(self, w):
        w.put_int(self.size)

    def write(self, w, val):
        if val is None:
            w.put_int(-1)
        else:
            if w._tds._tds._login.bytes_to_unicode:
                val = tds_base.force_unicode(val)
            if isinstance(val, six.text_type):
                val, _ = self._codec.encode(val)
            w.put_int(len(val))
            w.write(val)

    def read(self, r):
        size = r.get_byte()
        if size == 0:
            return None
        tds_base.readall(r, size)  # textptr
        tds_base.readall(r, 8)  # timestamp
        colsize = r.get_int()
        if self._chunk_handler is None:
            if r._session._tds._login.bytes_to_unicode:
                self._chunk_handler = _DefaultChunkedHandler(StringIO())
            else:
                self._chunk_handler = _DefaultChunkedHandler(BytesIO())
        if r._session._tds._login.bytes_to_unicode:
            for chunk in tds_base.iterdecode(read_chunks(r, colsize), self._codec):
                self._chunk_handler.add_chunk(chunk)
        else:
            for chunk in read_chunks(r, colsize):
                self._chunk_handler.add_chunk(chunk)
        return self._chunk_handler.end()

    def set_chunk_handler(self, chunk_handler):
        self._chunk_handler = chunk_handler


class Text71Serializer(Text70Serializer):
    def __repr__(self):
        return 'Text71(size={}, table_name={}, collation={})'.format(
            self.size, self._table_name, repr(self._collation)
        )

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        collation = r.get_collation()
        table_name = r.read_ucs2(r.get_smallint())
        return cls(size, table_name, collation)

    def write_info(self, w):
        w.put_int(self.size)
        w.put_collation(self._collation)


class Text72Serializer(Text71Serializer):
    def __init__(self, size=0, table_name_parts=(), collation=raw_collation):
        super(Text72Serializer, self).__init__(size=size, table_name='.'.join(table_name_parts), collation=collation)
        self._table_name_parts = table_name_parts

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        collation = r.get_collation()
        num_parts = r.get_byte()
        parts = []
        for _ in range(num_parts):
            parts.append(r.read_ucs2(r.get_smallint()))
        return cls(size, parts, collation)


class NText70Serializer(BaseTypeSerializer):
    type = tds_base.SYBNTEXT
    declaration = 'NTEXT'

    def __init__(self, size=0, table_name='', collation=raw_collation):
        super(NText70Serializer, self).__init__(size=size)
        self._collation = collation
        self._table_name = table_name
        self._chunk_handler = _DefaultChunkedHandler(StringIO())

    def __repr__(self):
        return 'NText70(size={}, table_name={})'.format(self.size, self._table_name)

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        table_name = r.read_ucs2(r.get_smallint())
        return cls(size, table_name)

    def read(self, r):
        textptr_size = r.get_byte()
        if textptr_size == 0:
            return None
        tds_base.readall(r, textptr_size)  # textptr
        tds_base.readall(r, 8)  # timestamp
        colsize = r.get_int()
        for chunk in tds_base.iterdecode(read_chunks(r, colsize), ucs2_codec):
            self._chunk_handler.add_chunk(chunk)
        return self._chunk_handler.end()

    def write_info(self, w):
        w.put_int(self.size * 2)

    def write(self, w, val):
        if val is None:
            w.put_int(-1)
        else:
            w.put_int(len(val) * 2)
            w.write_ucs2(val)

    def set_chunk_handler(self, chunk_handler):
        self._chunk_handler = chunk_handler


class NText71Serializer(NText70Serializer):
    def __repr__(self):
        return 'NText71(size={}, table_name={}, collation={})'.format(self.size,
                                                                      self._table_name,
                                                                      repr(self._collation))

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        collation = r.get_collation()
        table_name = r.read_ucs2(r.get_smallint())
        return cls(size, table_name, collation)

    def write_info(self, w):
        w.put_int(self.size)
        w.put_collation(self._collation)


class NText72Serializer(NText71Serializer):
    def __init__(self, size=0, table_name_parts=(), collation=raw_collation):
        super(NText72Serializer, self).__init__(size=size, collation=collation)
        self._table_name_parts = table_name_parts

    def __repr__(self):
        return 'NText72Serializer(s={},table_name={},coll={})'.format(
            self.size, self._table_name_parts, self._collation)

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        collation = r.get_collation()
        num_parts = r.get_byte()
        parts = []
        for _ in range(num_parts):
            parts.append(r.read_ucs2(r.get_smallint()))
        return cls(size, parts, collation)


class Binary(bytes, SqlValueMetaclass):
    def __repr__(self):
        return 'Binary({0})'.format(super(Binary, self).__repr__())


class VarBinarySerializer(BaseTypeSerializer):
    type = tds_base.XSYBVARBINARY

    def __init__(self, size):
        super(VarBinarySerializer, self).__init__(size=size)

    def __repr__(self):
        return 'VarBinary({})'.format(self.size)

    @classmethod
    def from_stream(cls, r):
        size = r.get_usmallint()
        return cls(size)

    def write_info(self, w):
        w.put_usmallint(self.size)

    def write(self, w, val):
        if val is None:
            w.put_usmallint(0xffff)
        else:
            w.put_usmallint(len(val))
            w.write(val)

    def read(self, r):
        size = r.get_usmallint()
        if size == 0xffff:
            return None
        return tds_base.readall(r, size)


class VarBinarySerializer72(VarBinarySerializer):
    def __repr__(self):
        return 'VarBinary72({})'.format(self.size)

    @classmethod
    def from_stream(cls, r):
        size = r.get_usmallint()
        if size == 0xffff:
            return VarBinarySerializerMax()
        return cls(size)


class VarBinarySerializerMax(VarBinarySerializer):
    def __init__(self):
        super(VarBinarySerializerMax, self).__init__(0)
        self._chunk_handler = _DefaultChunkedHandler(BytesIO())

    def __repr__(self):
        return 'VarBinaryMax()'

    def write_info(self, w):
        w.put_usmallint(tds_base.PLP_MARKER)

    def write(self, w, val):
        if val is None:
            w.put_uint8(tds_base.PLP_NULL)
        else:
            w.put_uint8(len(val))
            if val:
                w.put_uint(len(val))
                w.write(val)
            w.put_uint(0)

    def read(self, r):
        r = PlpReader(r)
        if r.is_null():
            return None
        for chunk in r.chunks():
            self._chunk_handler.add_chunk(chunk)
        return self._chunk_handler.end()

    def set_chunk_handler(self, chunk_handler):
        self._chunk_handler = chunk_handler


class UDT72Serializer(BaseTypeSerializer):
    # Data type definition stream used for UDT_INFO in TYPE_INFO
    # https://msdn.microsoft.com/en-us/library/a57df60e-d0a6-4e7e-a2e5-ccacd277c673/
    def __init__(self, max_byte_size, db_name, schema_name, type_name,
                 assembly_qualified_name):
        self.max_byte_size = max_byte_size
        self.db_name = db_name
        self.schema_name = schema_name
        self.type_name = type_name
        self.assembly_qualified_name = assembly_qualified_name
        super(UDT72Serializer, self).__init__()

    def __repr__(self):
        return ('UDT72Serializer(max_byte_size={}, db_name={}, '
                'schema_name={}, type_name={}, '
                'assembly_qualified_name={})'.format(
                    *map(repr, (
                        self.max_byte_size, self.db_name, self.schema_name,
                        self.type_name, self.assembly_qualified_name)))
        )

    @classmethod
    def from_stream(cls, r):
        # MAX_BYTE_SIZE
        max_byte_size = r.get_usmallint()
        assert max_byte_size == 0xffff or 1 < max_byte_size < 8000
        # DB_NAME -- B_VARCHAR
        db_name = r.read_ucs2(r.get_byte())
        # SCHEMA_NAME -- B_VARCHAR
        schema_name = r.read_ucs2(r.get_byte())
        # TYPE_NAME -- B_VARCHAR
        type_name = r.read_ucs2(r.get_byte())
        # UDT_METADATA --
        # a US_VARCHAR (2 bytes length prefix)
        # containing ASSEMBLY_QUALIFIED_NAME
        assembly_qualified_name = r.read_ucs2(r.get_smallint())
        return cls(max_byte_size, db_name, schema_name, type_name,
                   assembly_qualified_name)

    def read(self, r):
        r = PlpReader(r)
        if r.is_null():
            return None
        return b''.join(r.chunks())


class UDT72SerializerMax(UDT72Serializer):
    def __init__(self, *args, **kwargs):
        super(UDT72SerializerMax, self).__init__(0, *args, **kwargs)


class Image70Serializer(BaseTypeSerializer):
    type = tds_base.SYBIMAGE
    declaration = 'IMAGE'

    def __init__(self, size=0, table_name=''):
        super(Image70Serializer, self).__init__(size=size)
        self._table_name = table_name
        self._chunk_handler = _DefaultChunkedHandler(BytesIO())

    def __repr__(self):
        return 'Image70(tn={},s={})'.format(repr(self._table_name), self.size)

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        table_name = r.read_ucs2(r.get_smallint())
        return cls(size, table_name)

    def read(self, r):
        size = r.get_byte()
        if size == 16:  # Jeff's hack
            tds_base.readall(r, 16)  # textptr
            tds_base.readall(r, 8)  # timestamp
            colsize = r.get_int()
            for chunk in read_chunks(r, colsize):
                self._chunk_handler.add_chunk(chunk)
            return self._chunk_handler.end()
        else:
            return None

    def write(self, w, val):
        if val is None:
            w.put_int(-1)
            return
        w.put_int(len(val))
        w.write(val)

    def write_info(self, w):
        w.put_int(self.size)

    def set_chunk_handler(self, chunk_handler):
        self._chunk_handler = chunk_handler


class Image72Serializer(Image70Serializer):
    def __init__(self, size=0, parts=()):
        super(Image72Serializer, self).__init__(size=size, table_name='.'.join(parts))
        self._parts = parts

    def __repr__(self):
        return 'Image72(p={},s={})'.format(self._parts, self.size)

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        num_parts = r.get_byte()
        parts = []
        for _ in range(num_parts):
            parts.append(r.read_ucs2(r.get_usmallint()))
        return Image72Serializer(size, parts)


_datetime_base_date = datetime.datetime(1900, 1, 1)


class SmallDateTimeType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'SMALLDATETIME'


class DateTimeType(SqlTypeMetaclass):
    def get_declaration(self):
        return 'DATETIME'


class SmallDateTime(SqlValueMetaclass):
    """Corresponds to MSSQL smalldatetime"""
    def __init__(self, days, minutes):
        """

        @param days: Days since 1900-01-01
        @param minutes: Minutes since 00:00:00
        """
        self._days = days
        self._minutes = minutes

    @property
    def days(self):
        return self._days

    @property
    def minutes(self):
        return self._minutes

    def to_pydatetime(self):
        return _datetime_base_date + datetime.timedelta(days=self._days, minutes=self._minutes)

    @classmethod
    def from_pydatetime(cls, dt):
        days = (dt - _datetime_base_date).days
        minutes = dt.hour * 60 + dt.minute
        return cls(days=days, minutes=minutes)


class BaseDateTimeSerializer(BaseTypeSerializer):
    def write(self, w, value):
        raise NotImplementedError

    def write_info(self, w):
        raise NotImplementedError

    def read(self, r):
        raise NotImplementedError

    @classmethod
    def from_stream(cls, r):
        raise NotImplementedError


class SmallDateTimeSerializer(BasePrimitiveTypeSerializer, BaseDateTimeSerializer):
    type = tds_base.SYBDATETIME4
    declaration = 'SMALLDATETIME'

    _struct = struct.Struct('<HH')

    def write(self, w, val):
        if val.tzinfo:
            if not w.session.use_tz:
                raise tds_base.DataError('Timezone-aware datetime is used without specifying use_tz')
            val = val.astimezone(w.session.use_tz).replace(tzinfo=None)
        dt = SmallDateTime.from_pydatetime(val)
        w.pack(self._struct, dt.days, dt.minutes)

    def read(self, r):
        days, minutes = r.unpack(self._struct)
        dt = SmallDateTime(days=days, minutes=minutes)
        tzinfo = None
        if r.session.tzinfo_factory is not None:
            tzinfo = r.session.tzinfo_factory(0)
        return dt.to_pydatetime().replace(tzinfo=tzinfo)

SmallDateTimeSerializer.instance = SmallDateTimeSerializer()


class DateTime(SqlValueMetaclass):
    """Corresponds to MSSQL datetime"""
    MIN_PYDATETIME = datetime.datetime(1753, 1, 1, 0, 0, 0)
    MAX_PYDATETIME = datetime.datetime(9999, 12, 31, 23, 59, 59, 997000)

    def __init__(self, days, time_part):
        """

        @param days: Days since 1900-01-01
        @param time_part: Number of 1/300 of seconds since 00:00:00
        """
        self._days = days
        self._time_part = time_part

    @property
    def days(self):
        return self._days

    @property
    def time_part(self):
        return self._time_part

    def to_pydatetime(self):
        ms = int(round(self._time_part % 300 * 10 / 3.0))
        secs = self._time_part // 300
        return _datetime_base_date + datetime.timedelta(days=self._days, seconds=secs, milliseconds=ms)

    @classmethod
    def from_pydatetime(cls, dt):
        if not (cls.MIN_PYDATETIME <= dt <= cls.MAX_PYDATETIME):
            raise tds_base.DataError('Datetime is out of range')
        days = (dt - _datetime_base_date).days
        ms = dt.microsecond // 1000
        tm = (dt.hour * 60 * 60 + dt.minute * 60 + dt.second) * 300 + int(round(ms * 3 / 10.0))
        return cls(days=days, time_part=tm)


class DateTimeSerializer(BasePrimitiveTypeSerializer, BaseDateTimeSerializer):
    type = tds_base.SYBDATETIME
    declaration = 'DATETIME'

    _struct = struct.Struct('<ll')

    def write(self, w, val):
        if val.tzinfo:
            if not w.session.use_tz:
                raise tds_base.DataError('Timezone-aware datetime is used without specifying use_tz')
            val = val.astimezone(w.session.use_tz).replace(tzinfo=None)
        w.write(self.encode(val))

    def read(self, r):
        days, t = r.unpack(self._struct)
        tzinfo = None
        if r.session.tzinfo_factory is not None:
            tzinfo = r.session.tzinfo_factory(0)
        return _applytz(self.decode(days, t), tzinfo)

    @classmethod
    def encode(cls, value):
        if type(value) == datetime.date:
            value = datetime.datetime.combine(value, datetime.time(0, 0, 0))
        dt = DateTime.from_pydatetime(value)
        return cls._struct.pack(dt.days, dt.time_part)

    @classmethod
    def decode(cls, days, time_part):
        dt = DateTime(days=days, time_part=time_part)
        return dt.to_pydatetime()

DateTimeSerializer.instance = DateTimeSerializer()


class DateTimeNSerializer(BaseTypeSerializerN, BaseDateTimeSerializer):
    type = tds_base.SYBDATETIMN
    subtypes = {
        4: SmallDateTimeSerializer.instance,
        8: DateTimeSerializer.instance,
    }


_datetime2_base_date = datetime.datetime(1, 1, 1)


class DateType(SqlTypeMetaclass):
    type = tds_base.SYBMSDATE

    def get_declaration(self):
        return "DATE"


class Date(SqlValueMetaclass):
    MIN_PYDATE = datetime.date(1, 1, 1)
    MAX_PYDATE = datetime.date(9999, 12, 31)

    def __init__(self, days):
        """
        Creates sql date object
        @param days: Days since 0001-01-01
        """
        self._days = days

    @property
    def days(self):
        return self._days

    def to_pydate(self):
        """
        Converts sql date to Python date
        @return: Python date
        """
        return (_datetime2_base_date + datetime.timedelta(days=self._days)).date()

    @classmethod
    def from_pydate(cls, pydate):
        """
        Creates sql date object from Python date object.
        @param pydate: Python date
        @return: sql date
        """
        return cls(days=(datetime.datetime.combine(pydate, datetime.time(0, 0, 0)) - _datetime2_base_date).days)


class TimeType(SqlTypeMetaclass):
    type = tds_base.SYBMSTIME

    def __init__(self, precision=7):
        self._precision = precision

    @property
    def precision(self):
        return self._precision

    def get_declaration(self):
        return 'TIME({0})'.format(self.precision)


class Time(SqlValueMetaclass):
    def __init__(self, nsec):
        """
        Creates sql time object.
        Maximum precision which sql server supports is 100 nanoseconds.
        Values more precise than 100 nanoseconds will be truncated.
        @param nsec: Nanoseconds from 00:00:00
        """
        self._nsec = nsec

    @property
    def nsec(self):
        return self._nsec

    def to_pytime(self):
        """
        Converts sql time object into Python's time object
        this will truncate nanoseconds to microseconds
        @return: naive time
        """
        nanoseconds = self._nsec
        hours = nanoseconds // 1000000000 // 60 // 60
        nanoseconds -= hours * 60 * 60 * 1000000000
        minutes = nanoseconds // 1000000000 // 60
        nanoseconds -= minutes * 60 * 1000000000
        seconds = nanoseconds // 1000000000
        nanoseconds -= seconds * 1000000000
        return datetime.time(hours, minutes, seconds, nanoseconds // 1000)

    @classmethod
    def from_pytime(cls, pytime):
        """
        Converts Python time object to sql time object
        ignoring timezone
        @param pytime: Python time object
        @return: sql time object
        """
        secs = pytime.hour * 60 * 60 + pytime.minute * 60 + pytime.second
        nsec = secs * 10 ** 9 + pytime.microsecond * 1000
        return cls(nsec=nsec)


class DateTime2Type(SqlTypeMetaclass):
    type = tds_base.SYBMSDATETIME2

    def __init__(self, precision=7):
        self._precision = precision

    @property
    def precision(self):
        return self._precision

    def get_declaration(self):
        return 'DATETIME2({0})'.format(self.precision)


class DateTime2(SqlValueMetaclass):
    type = tds_base.SYBMSDATETIME2

    def __init__(self, date, time):
        """
        Creates datetime2 object
        @param date: sql date object
        @param time: sql time object
        """
        self._date = date
        self._time = time

    @property
    def date(self):
        return self._date

    @property
    def time(self):
        return self._time

    def to_pydatetime(self):
        """
        Converts datetime2 object into Python's datetime.datetime object
        @return: naive datetime.datetime
        """
        return datetime.datetime.combine(self._date.to_pydate(), self._time.to_pytime())

    @classmethod
    def from_pydatetime(cls, pydatetime):
        """
        Creates sql datetime2 object from Python datetime object
        ignoring timezone
        @param pydatetime: Python datetime object
        @return: sql datetime2 object
        """
        return cls(date=Date.from_pydate(pydatetime.date),
                   time=Time.from_pytime(pydatetime.time))


class DateTimeOffsetType(SqlTypeMetaclass):
    type = tds_base.SYBMSDATETIMEOFFSET

    def __init__(self, precision=7):
        self._precision = precision

    @property
    def precision(self):
        return self._precision

    def get_declaration(self):
        return 'DATETIMEOFFSET({0})'.format(self.precision)


class DateTimeOffset(SqlValueMetaclass):
    def __init__(self, date, time, offset):
        """
        Creates datetime2 object
        @param date: sql date object in UTC
        @param time: sql time object in UTC
        @param offset: time zone offset in minutes
        """
        self._date = date
        self._time = time
        self._offset = offset

    def to_pydatetime(self):
        """
        Converts datetimeoffset object into Python's datetime.datetime object
        @return: time zone aware datetime.datetime
        """
        dt = datetime.datetime.combine(self._date.to_pydate(), self._time.to_pytime())
        from .tz import FixedOffsetTimezone
        return dt.replace(tzinfo=_utc).astimezone(FixedOffsetTimezone(self._offset))


class BaseDateTime73Serializer(BaseTypeSerializer):
    def write(self, w, value):
        raise NotImplementedError

    def write_info(self, w):
        raise NotImplementedError

    def read(self, r):
        raise NotImplementedError

    @classmethod
    def from_stream(cls, r):
        raise NotImplementedError

    _precision_to_len = {
        0: 3,
        1: 3,
        2: 3,
        3: 4,
        4: 4,
        5: 5,
        6: 5,
        7: 5,
    }

    def _write_time(self, w, t, prec):
        val = t.nsec // (10 ** (9 - prec))
        w.write(struct.pack('<Q', val)[:self._precision_to_len[prec]])

    @staticmethod
    def _read_time(r, size, prec):
        time_buf = tds_base.readall(r, size)
        val = _decode_num(time_buf)
        val *= 10 ** (7 - prec)
        nanoseconds = val * 100
        return Time(nsec=nanoseconds)

    @staticmethod
    def _write_date(w, value):
        days = value.days
        buf = struct.pack('<l', days)[:3]
        w.write(buf)

    @staticmethod
    def _read_date(r):
        days = _decode_num(tds_base.readall(r, 3))
        return Date(days=days)


class MsDateSerializer(BasePrimitiveTypeSerializer, BaseDateTime73Serializer):
    type = tds_base.SYBMSDATE
    declaration = 'DATE'

    def __init__(self, typ):
        super(MsDateSerializer, self).__init__()
        self._typ = typ

    @classmethod
    def from_stream(cls, r):
        return cls(DateType())

    def write(self, w, value):
        if value is None:
            w.put_byte(0)
        else:
            w.put_byte(3)
            self._write_date(w, Date.from_pydate(value))

    def read_fixed(self, r):
        return self._read_date(r).to_pydate()

    def read(self, r):
        size = r.get_byte()
        if size == 0:
            return None
        return self._read_date(r).to_pydate()


class MsTimeSerializer(BaseDateTime73Serializer):
    type = tds_base.SYBMSTIME

    def __init__(self, typ):
        super(MsTimeSerializer, self).__init__(precision=typ.precision, size=self._precision_to_len[typ.precision])
        self._typ = typ

    @classmethod
    def read_type(cls, r):
        prec = r.get_byte()
        return TimeType(precision=prec)

    @classmethod
    def from_stream(cls, r):
        return cls(cls.read_type(r))

    def write_info(self, w):
        w.put_byte(self._typ.precision)

    def write(self, w, value):
        if value is None:
            w.put_byte(0)
        else:
            if value.tzinfo:
                if not w.session.use_tz:
                    raise tds_base.DataError('Timezone-aware datetime is used without specifying use_tz')
                value = value.astimezone(w.session.use_tz).replace(tzinfo=None)
            w.put_byte(self.size)
            self._write_time(w, Time.from_pytime(value), self._typ.precision)

    def read_fixed(self, r, size):
        res = self._read_time(r, size, self._typ.precision).to_pytime()
        if r.session.tzinfo_factory is not None:
            tzinfo = r.session.tzinfo_factory(0)
            res = res.replace(tzinfo=tzinfo)
        return res

    def read(self, r):
        size = r.get_byte()
        if size == 0:
            return None
        return self.read_fixed(r, size)


class DateTime2Serializer(BaseDateTime73Serializer):
    type = tds_base.SYBMSDATETIME2

    def __init__(self, typ):
        super(DateTime2Serializer, self).__init__(precision=typ.precision,
                                                  size=self._precision_to_len[typ.precision] + 3)
        self._typ = typ

    @classmethod
    def from_stream(cls, r):
        prec = r.get_byte()
        return cls(DateTime2Type(precision=prec))

    def write_info(self, w):
        w.put_byte(self._typ.precision)

    def write(self, w, value):
        if value is None:
            w.put_byte(0)
        else:
            if value.tzinfo:
                if not w.session.use_tz:
                    raise tds_base.DataError('Timezone-aware datetime is used without specifying use_tz')
                value = value.astimezone(w.session.use_tz).replace(tzinfo=None)
            w.put_byte(self.size)
            self._write_time(w, Time.from_pytime(value), self._typ.precision)
            self._write_date(w, Date.from_pydate(value))

    def read_fixed(self, r, size):
        time = self._read_time(r, size - 3, self._typ.precision)
        date = self._read_date(r)
        dt = DateTime2(date=date, time=time)
        res = dt.to_pydatetime()
        if r.session.tzinfo_factory is not None:
            tzinfo = r.session.tzinfo_factory(0)
            res = res.replace(tzinfo=tzinfo)
        return res

    def read(self, r):
        size = r.get_byte()
        if size == 0:
            return None
        return self.read_fixed(r, size)


class DateTimeOffsetSerializer(BaseDateTime73Serializer):
    type = tds_base.SYBMSDATETIMEOFFSET

    def __init__(self, typ):
        super(DateTimeOffsetSerializer, self).__init__(precision=typ.precision,
                                                       size=self._precision_to_len[typ.precision] + 5)
        self._typ = typ

    @classmethod
    def from_stream(cls, r):
        prec = r.get_byte()
        return cls(DateTimeOffsetType(precision=prec))

    def write_info(self, w):
        w.put_byte(self._typ.precision)

    def write(self, w, value):
        if value is None:
            w.put_byte(0)
        else:
            utcoffset = value.utcoffset()
            value = value.astimezone(_utc).replace(tzinfo=None)

            w.put_byte(self.size)
            self._write_time(w, Time.from_pytime(value), self._typ.precision)
            self._write_date(w, Date.from_pydate(value))
            w.put_smallint(int(tds_base.total_seconds(utcoffset)) // 60)

    def read_fixed(self, r, size):
        time = self._read_time(r, size - 5, self._typ.precision)
        date = self._read_date(r)
        offset = r.get_smallint()
        dt = DateTimeOffset(date=date, time=time, offset=offset)
        return dt.to_pydatetime()

    def read(self, r):
        size = r.get_byte()
        if size == 0:
            return None
        return self.read_fixed(r, size)


class MsDecimalSerializer(BaseTypeSerializer):
    type = tds_base.SYBDECIMAL

    _max_size = 17

    _bytes_per_prec = [
        #
        # precision can't be 0 but using a value > 0 assure no
        # core if for some bug it's 0...
        #
        1,
        5, 5, 5, 5, 5, 5, 5, 5, 5,
        9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
        13, 13, 13, 13, 13, 13, 13, 13, 13,
        17, 17, 17, 17, 17, 17, 17, 17, 17, 17,
    ]

    _info_struct = struct.Struct('BBB')

    def __init__(self, precision=18, scale=0):
        super(MsDecimalSerializer, self).__init__(precision=precision,
                                                  scale=scale,
                                                  size=self._bytes_per_prec[precision])
        if precision > 38:
            raise tds_base.DataError('Precision of decimal value is out of range')

    def __repr__(self):
        return 'MsDecimal(scale={}, prec={})'.format(self.scale, self.precision)

    @classmethod
    def from_value(cls, value):
        sql_type = DecimalType.from_value(value)
        return cls(scale=sql_type.scale, prec=sql_type.precision)

    @classmethod
    def from_stream(cls, r):
        size, prec, scale = r.unpack(cls._info_struct)
        return cls(scale=scale, precision=prec)

    def write_info(self, w):
        w.pack(self._info_struct, self.size, self.precision, self.scale)

    def write(self, w, value):
        with decimal.localcontext() as context:
            context.prec = 38
            if value is None:
                w.put_byte(0)
                return
            if not isinstance(value, decimal.Decimal):
                value = decimal.Decimal(value)
            value = value.normalize()
            scale = self.scale
            size = self.size
            w.put_byte(size)
            val = value
            positive = 1 if val > 0 else 0
            w.put_byte(positive)  # sign
            if not positive:
                val *= -1
            size -= 1
            val *= 10 ** scale
            for i in range(size):
                w.put_byte(int(val % 256))
                val //= 256
            assert val == 0

    def _decode(self, positive, buf):
        val = _decode_num(buf)
        val = decimal.Decimal(val)
        with decimal.localcontext() as ctx:
            ctx.prec = 38
            if not positive:
                val *= -1
            val /= 10 ** self._scale
        return val

    def read_fixed(self, r, size):
        positive = r.get_byte()
        buf = tds_base.readall(r, size - 1)
        return self._decode(positive, buf)

    def read(self, r):
        size = r.get_byte()
        if size <= 0:
            return None
        return self.read_fixed(r, size)


class Money4Serializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBMONEY4
    declaration = 'SMALLMONEY'

    def read(self, r):
        return decimal.Decimal(r.get_int()) / 10000

    def write(self, w, val):
        val = int(val * 10000)
        w.put_int(val)

Money4Serializer.instance = Money4Serializer()


class Money8Serializer(BasePrimitiveTypeSerializer):
    type = tds_base.SYBMONEY
    declaration = 'MONEY'

    _struct = struct.Struct('<lL')

    def read(self, r):
        hi, lo = r.unpack(self._struct)
        val = hi * (2 ** 32) + lo
        return decimal.Decimal(val) / 10000

    def write(self, w, val):
        val *= 10000
        hi = int(val // (2 ** 32))
        lo = int(val % (2 ** 32))
        w.pack(self._struct, hi, lo)

Money8Serializer.instance = Money8Serializer()


class MoneyNSerializer(BaseTypeSerializerN):
    type = tds_base.SYBMONEYN

    subtypes = {
        4: Money4Serializer.instance,
        8: Money8Serializer.instance,
    }


class MsUniqueSerializer(BaseTypeSerializer):
    type = tds_base.SYBUNIQUE
    declaration = 'UNIQUEIDENTIFIER'
    instance = None

    def __repr__(self):
        return 'MsUniqueSerializer()'

    @classmethod
    def from_stream(cls, r):
        size = r.get_byte()
        if size != 16:
            raise tds_base.InterfaceError('Invalid size of UNIQUEIDENTIFIER field')
        return cls.instance

    def write_info(self, w):
        w.put_byte(16)

    def write(self, w, value):
        if value is None:
            w.put_byte(0)
        else:
            w.put_byte(16)
            w.write(value.bytes_le)

    @staticmethod
    def read_fixed(r, size):
        return uuid.UUID(bytes_le=tds_base.readall(r, size))

    def read(self, r):
        size = r.get_byte()
        if size == 0:
            return None
        if size != 16:
            raise tds_base.InterfaceError('Invalid size of UNIQUEIDENTIFIER field')
        return self.read_fixed(r, size)
MsUniqueSerializer.instance = MsUniqueSerializer()


def _variant_read_str(r, size):
    collation = r.get_collation()
    r.get_usmallint()
    return r.read_str(size, collation.get_codec())


def _variant_read_nstr(r, size):
    r.get_collation()
    r.get_usmallint()
    return r.read_str(size, ucs2_codec)


def _variant_read_decimal(r, size):
    prec, scale = r.unpack(VariantSerializer.decimal_info_struct)
    return MsDecimalSerializer(precision=prec, scale=scale).read_fixed(r, size)


def _variant_read_binary(r, size):
    r.get_usmallint()
    return tds_base.readall(r, size)


class VariantSerializer(BaseTypeSerializer):
    type = tds_base.SYBVARIANT
    declaration = 'SQL_VARIANT'

    decimal_info_struct = struct.Struct('BB')

    _type_map = {
        tds_base.GUIDTYPE: lambda r, size: MsUniqueSerializer.instance.read_fixed(r, size),
        tds_base.BITTYPE: lambda r, size: BitSerializer.instance.read(r),
        tds_base.INT1TYPE: lambda r, size: TinyIntSerializer.instance.read(r),
        tds_base.INT2TYPE: lambda r, size: SmallIntSerializer.instance.read(r),
        tds_base.INT4TYPE: lambda r, size: IntSerializer.instance.read(r),
        tds_base.INT8TYPE: lambda r, size: BigIntSerializer.instance.read(r),
        tds_base.DATETIMETYPE: lambda r, size: DateTimeSerializer.instance.read(r),
        tds_base.DATETIM4TYPE: lambda r, size: SmallDateTimeSerializer.instance.read(r),
        tds_base.FLT4TYPE: lambda r, size: RealSerializer.instance.read(r),
        tds_base.FLT8TYPE: lambda r, size: FloatSerializer.instance.read(r),
        tds_base.MONEYTYPE: lambda r, size: Money8Serializer.instance.read(r),
        tds_base.MONEY4TYPE: lambda r, size: Money4Serializer.instance.read(r),
        tds_base.DATENTYPE: lambda r, size: MsDateSerializer(DateType()).read_fixed(r),

        tds_base.TIMENTYPE: lambda r, size: MsTimeSerializer(TimeType(precision=r.get_byte())).read_fixed(r, size),
        tds_base.DATETIME2NTYPE: lambda r, size: DateTime2Serializer(
            DateTime2Type(precision=r.get_byte())).read_fixed(r, size),
        tds_base.DATETIMEOFFSETNTYPE: lambda r, size: DateTimeOffsetSerializer(
            DateTimeOffsetType(precision=r.get_byte())).read_fixed(r, size),

        tds_base.BIGVARBINTYPE: _variant_read_binary,
        tds_base.BIGBINARYTYPE: _variant_read_binary,

        tds_base.NUMERICNTYPE: _variant_read_decimal,
        tds_base.DECIMALNTYPE: _variant_read_decimal,

        tds_base.BIGVARCHRTYPE: _variant_read_str,
        tds_base.BIGCHARTYPE: _variant_read_str,
        tds_base.NVARCHARTYPE: _variant_read_nstr,
        tds_base.NCHARTYPE: _variant_read_nstr,

    }

    @classmethod
    def from_stream(cls, r):
        size = r.get_int()
        return VariantSerializer(size)

    def write_info(self, w):
        w.put_int(self.size)

    def read(self, r):
        size = r.get_int()
        if size == 0:
            return None

        type_id = r.get_byte()
        prop_bytes = r.get_byte()
        type_factory = self._type_map.get(type_id)
        if not type_factory:
            r.session.bad_stream('Variant type invalid', type_id)
        return type_factory(r, size - prop_bytes - 2)

    def write(self, w, val):
        if val is None:
            w.put_int(0)
            return
        raise NotImplementedError


class TableType(SqlTypeMetaclass):
    """
    Used to serialize table valued parameters

    spec: https://msdn.microsoft.com/en-us/library/dd304813.aspx
    """
    def __init__(self, typ_schema, typ_name, columns):
        """
        @param typ_schema: Schema where TVP type defined
        @param typ_name: Name of TVP type
        @param columns: List of column types
        """
        if len(typ_schema) > 128:
            raise ValueError("Schema part of TVP name should be no longer than 128 characters")
        if len(typ_name) > 128:
            raise ValueError("Name part of TVP name should be no longer than 128 characters")
        if columns is not None:
            if len(columns) > 1024:
                raise ValueError("TVP cannot have more than 1024 columns")
            if len(columns) < 1:
                raise ValueError("TVP must have at least one column")
        self._typ_dbname = ''  # dbname should always be empty string for TVP according to spec
        self._typ_schema = typ_schema
        self._typ_name = typ_name
        self._columns = columns

    def __repr__(self):
        return 'TableType(s={},n={},cols={})'.format(
            self._typ_schema, self._typ_name, repr(self._columns)
        )

    def get_declaration(self):
        assert not self._typ_dbname
        if self._typ_schema:
            full_name = '{}.{}'.format(self._typ_schema, self._typ_name)
        else:
            full_name = self._typ_name
        return '{} READONLY'.format(full_name)

    @property
    def typ_schema(self):
        return self._typ_schema

    @property
    def typ_name(self):
        return self._typ_name

    @property
    def columns(self):
        return self._columns


class TableValuedParam(SqlValueMetaclass):
    """
    Used to represent a value of table-valued parameter
    """
    def __init__(self, type_name=None, columns=None, rows=None):
        # parsing type name
        self._typ_schema = ''
        self._typ_name = ''
        if type_name:
            parts = type_name.split('.')
            if len(parts) > 2:
                raise ValueError('Type name should consist of at most 2 parts, e.g. dbo.MyType')
            self._typ_name = parts[-1]
            if len(parts) > 1:
                self._typ_schema = parts[0]

        self._columns = columns
        self._rows = rows

    @property
    def typ_name(self):
        return self._typ_name

    @property
    def typ_schema(self):
        return self._typ_schema

    @property
    def columns(self):
        return self._columns

    @property
    def rows(self):
        return self._rows

    def is_null(self):
        return self._rows is None

    def peek_row(self):
        try:
            rows = iter(self._rows)
        except TypeError:
            raise tds_base.DataError('rows should be iterable')

        try:
            row = next(rows)
        except StopIteration:
            # no rows
            raise tds_base.DataError("Cannot infer columns from rows for TVP because there are no rows")
        else:
            # put row back
            self._rows = itertools.chain([row], rows)
        return row


class TableSerializer(BaseTypeSerializer):
    """
    Used to serialize table valued parameters

    spec: https://msdn.microsoft.com/en-us/library/dd304813.aspx
    """

    type = tds_base.TVPTYPE

    def read(self, r):
        """ According to spec TDS does not support output TVP values """
        raise NotImplementedError

    @classmethod
    def from_stream(cls, r):
        """ According to spec TDS does not support output TVP values """
        raise NotImplementedError

    def __init__(self, table_type, columns_serializers):
        super(TableSerializer, self).__init__()
        self._table_type = table_type
        self._columns_serializers = columns_serializers

    @property
    def table_type(self):
        return self._table_type

    def __repr__(self):
        return 'TableSerializer(t={},c={})'.format(
            repr(self._table_type), repr(self._columns_serializers)
        )

    def write_info(self, w):
        """
        Writes TVP_TYPENAME structure

        spec: https://msdn.microsoft.com/en-us/library/dd302994.aspx
        @param w: TdsWriter
        @return:
        """
        w.write_b_varchar("")  # db_name, should be empty
        w.write_b_varchar(self._table_type.typ_schema)
        w.write_b_varchar(self._table_type.typ_name)

    def write(self, w, val):
        """
        Writes remaining part of TVP_TYPE_INFO structure, resuming from TVP_COLMETADATA

        specs:
        https://msdn.microsoft.com/en-us/library/dd302994.aspx
        https://msdn.microsoft.com/en-us/library/dd305261.aspx
        https://msdn.microsoft.com/en-us/library/dd303230.aspx

        @param w: TdsWriter
        @param val: TableValuedParam or None
        @return:
        """
        if val.is_null():
            w.put_usmallint(tds_base.TVP_NULL_TOKEN)
        else:
            columns = self._table_type.columns
            w.put_usmallint(len(columns))
            for i, column in enumerate(columns):
                w.put_uint(column.column_usertype)

                w.put_usmallint(column.flags)

                # TYPE_INFO structure: https://msdn.microsoft.com/en-us/library/dd358284.aspx

                serializer = self._columns_serializers[i]
                type_id = serializer.type
                w.put_byte(type_id)
                serializer.write_info(w)

                w.write_b_varchar('')  # ColName, must be empty in TVP according to spec

        # here can optionally send TVP_ORDER_UNIQUE and TVP_COLUMN_ORDERING
        # https://msdn.microsoft.com/en-us/library/dd305261.aspx

        # terminating optional metadata
        w.put_byte(tds_base.TVP_END_TOKEN)

        # now sending rows using TVP_ROW
        # https://msdn.microsoft.com/en-us/library/dd305261.aspx
        if val.rows:
            for row in val.rows:
                w.put_byte(tds_base.TVP_ROW_TOKEN)
                for i, col in enumerate(self._table_type.columns):
                    if not col.flags & tds_base.TVP_COLUMN_DEFAULT_FLAG:
                        self._columns_serializers[i].write(w, row[i])

        # terminating rows
        w.put_byte(tds_base.TVP_END_TOKEN)


_type_map = {
    tds_base.SYBINT1: TinyIntSerializer,
    tds_base.SYBINT2: SmallIntSerializer,
    tds_base.SYBINT4: IntSerializer,
    tds_base.SYBINT8: BigIntSerializer,
    tds_base.SYBINTN: IntNSerializer,
    tds_base.SYBBIT: BitSerializer,
    tds_base.SYBBITN: BitNSerializer,
    tds_base.SYBREAL: RealSerializer,
    tds_base.SYBFLT8: FloatSerializer,
    tds_base.SYBFLTN: FloatNSerializer,
    tds_base.SYBMONEY4: Money4Serializer,
    tds_base.SYBMONEY: Money8Serializer,
    tds_base.SYBMONEYN: MoneyNSerializer,
    tds_base.XSYBCHAR: VarChar70Serializer,
    tds_base.XSYBVARCHAR: VarChar70Serializer,
    tds_base.XSYBNCHAR: NVarChar70Serializer,
    tds_base.XSYBNVARCHAR: NVarChar70Serializer,
    tds_base.SYBTEXT: Text70Serializer,
    tds_base.SYBNTEXT: NText70Serializer,
    tds_base.SYBMSXML: XmlSerializer,
    tds_base.XSYBBINARY: VarBinarySerializer,
    tds_base.XSYBVARBINARY: VarBinarySerializer,
    tds_base.SYBIMAGE: Image70Serializer,
    tds_base.SYBNUMERIC: MsDecimalSerializer,
    tds_base.SYBDECIMAL: MsDecimalSerializer,
    tds_base.SYBVARIANT: VariantSerializer,
    tds_base.SYBMSDATE: MsDateSerializer,
    tds_base.SYBMSTIME: MsTimeSerializer,
    tds_base.SYBMSDATETIME2: DateTime2Serializer,
    tds_base.SYBMSDATETIMEOFFSET: DateTimeOffsetSerializer,
    tds_base.SYBDATETIME4: SmallDateTimeSerializer,
    tds_base.SYBDATETIME: DateTimeSerializer,
    tds_base.SYBDATETIMN: DateTimeNSerializer,
    tds_base.SYBUNIQUE: MsUniqueSerializer,
}

_type_map71 = _type_map.copy()
_type_map71.update({
    tds_base.XSYBCHAR: VarChar71Serializer,
    tds_base.XSYBNCHAR: NVarChar71Serializer,
    tds_base.XSYBVARCHAR: VarChar71Serializer,
    tds_base.XSYBNVARCHAR: NVarChar71Serializer,
    tds_base.SYBTEXT: Text71Serializer,
    tds_base.SYBNTEXT: NText71Serializer,
})

_type_map72 = _type_map.copy()
_type_map72.update({
    tds_base.XSYBCHAR: VarChar72Serializer,
    tds_base.XSYBNCHAR: NVarChar72Serializer,
    tds_base.XSYBVARCHAR: VarChar72Serializer,
    tds_base.XSYBNVARCHAR: NVarChar72Serializer,
    tds_base.SYBTEXT: Text72Serializer,
    tds_base.SYBNTEXT: NText72Serializer,
    tds_base.XSYBBINARY: VarBinarySerializer72,
    tds_base.XSYBVARBINARY: VarBinarySerializer72,
    tds_base.SYBIMAGE: Image72Serializer,
    tds_base.UDTTYPE: UDT72Serializer,
})

_type_map73 = _type_map72.copy()
_type_map73.update({
    tds_base.TVPTYPE: TableSerializer,
})


def sql_type_by_declaration(declaration):
    return _declarations_parser.parse(declaration)


class SerializerFactory(object):
    """
    Factory class for TDS data types
    """
    def __init__(self, tds_ver):
        self._tds_ver = tds_ver
        if self._tds_ver >= tds_base.TDS73:
            self._type_map = _type_map73
        elif self._tds_ver >= tds_base.TDS72:
            self._type_map = _type_map72
        elif self._tds_ver >= tds_base.TDS71:
            self._type_map = _type_map71
        else:
            self._type_map = _type_map

    def get_type_serializer(self, tds_type_id):
        type_class = self._type_map.get(tds_type_id)
        if not type_class:
            raise tds_base.InterfaceError('Invalid type id {}'.format(tds_type_id))
        return type_class

    def long_binary_type(self):
        if self._tds_ver >= tds_base.TDS72:
            return VarBinaryMaxType()
        else:
            return ImageType()

    def long_varchar_type(self):
        if self._tds_ver >= tds_base.TDS72:
            return VarCharMaxType()
        else:
            return TextType()

    def long_string_type(self):
        if self._tds_ver >= tds_base.TDS72:
            return NVarCharMaxType()
        else:
            return NTextType()

    def datetime(self, precision):
        if self._tds_ver >= tds_base.TDS72:
            return DateTime2Type(precision=precision)
        else:
            return DateTimeType()

    def has_datetime_with_tz(self):
        return self._tds_ver >= tds_base.TDS72

    def datetime_with_tz(self, precision):
        if self._tds_ver >= tds_base.TDS72:
            return DateTimeOffsetType(precision=precision)
        else:
            raise tds_base.DataError('Given TDS version does not support DATETIMEOFFSET type')

    def date(self):
        if self._tds_ver >= tds_base.TDS72:
            return DateType()
        else:
            return DateTimeType()

    def time(self, precision):
        if self._tds_ver >= tds_base.TDS72:
            return TimeType(precision=precision)
        else:
            raise tds_base.DataError('Given TDS version does not support TIME type')

    def serializer_by_declaration(self, declaration, connection):
        sql_type = sql_type_by_declaration(declaration)
        return self.serializer_by_type(sql_type=sql_type, collation=connection.collation)

    def serializer_by_type(self, sql_type, collation=raw_collation):
        typ = sql_type
        if isinstance(typ, BitType):
            return BitNSerializer(typ)
        elif isinstance(typ, TinyIntType):
            return IntNSerializer(typ)
        elif isinstance(typ, SmallIntType):
            return IntNSerializer(typ)
        elif isinstance(typ, IntType):
            return IntNSerializer(typ)
        elif isinstance(typ, BigIntType):
            return IntNSerializer(typ)
        elif isinstance(typ, RealType):
            return FloatNSerializer(size=4)
        elif isinstance(typ, FloatType):
            return FloatNSerializer(size=8)
        elif isinstance(typ, SmallMoneyType):
            return self._type_map[tds_base.SYBMONEYN](size=4)
        elif isinstance(typ, MoneyType):
            return self._type_map[tds_base.SYBMONEYN](size=8)
        elif isinstance(typ, CharType):
            return self._type_map[tds_base.XSYBCHAR](size=typ.size, collation=collation)
        elif isinstance(typ, VarCharType):
            return self._type_map[tds_base.XSYBVARCHAR](size=typ.size, collation=collation)
        elif isinstance(typ, VarCharMaxType):
            return VarCharMaxSerializer(collation=collation)
        elif isinstance(typ, NCharType):
            return self._type_map[tds_base.XSYBNCHAR](size=typ.size, collation=collation)
        elif isinstance(typ, NVarCharType):
            return self._type_map[tds_base.XSYBNVARCHAR](size=typ.size, collation=collation)
        elif isinstance(typ, NVarCharMaxType):
            return NVarCharMaxSerializer(collation=collation)
        elif isinstance(typ, TextType):
            return self._type_map[tds_base.SYBTEXT](collation=collation)
        elif isinstance(typ, NTextType):
            return self._type_map[tds_base.SYBNTEXT](collation=collation)
        elif isinstance(typ, XmlType):
            return self._type_map[tds_base.SYBMSXML]()
        elif isinstance(typ, BinaryType):
            return self._type_map[tds_base.XSYBBINARY]()
        elif isinstance(typ, VarBinaryType):
            return self._type_map[tds_base.XSYBVARBINARY](size=typ.size)
        elif isinstance(typ, VarBinaryMaxType):
            return VarBinarySerializerMax()
        elif isinstance(typ, ImageType):
            return self._type_map[tds_base.SYBIMAGE]()
        elif isinstance(typ, DecimalType):
            return self._type_map[tds_base.SYBDECIMAL](scale=typ.scale, precision=typ.precision)
        elif isinstance(typ, VariantType):
            return self._type_map[tds_base.SYBVARIANT](size=0)
        elif isinstance(typ, SmallDateTimeType):
            return self._type_map[tds_base.SYBDATETIMN](size=4)
        elif isinstance(typ, DateTimeType):
            return self._type_map[tds_base.SYBDATETIMN](size=8)
        elif isinstance(typ, DateType):
            return self._type_map[tds_base.SYBMSDATE](typ)
        elif isinstance(typ, TimeType):
            return self._type_map[tds_base.SYBMSTIME](typ)
        elif isinstance(typ, DateTime2Type):
            return self._type_map[tds_base.SYBMSDATETIME2](typ)
        elif isinstance(typ, DateTimeOffsetType):
            return self._type_map[tds_base.SYBMSDATETIMEOFFSET](typ)
        elif isinstance(typ, UniqueIdentifierType):
            return self._type_map[tds_base.SYBUNIQUE]()
        elif isinstance(typ, TableType):
            columns_serializers = None
            if typ.columns is not None:
                columns_serializers = [self.serializer_by_type(col.type) for col in typ.columns]
            return TableSerializer(table_type=typ, columns_serializers=columns_serializers)
        else:
            raise ValueError('Cannot map type {} to serializer.'.format(typ))


class DeclarationsParser(object):
    def __init__(self):
        declaration_parsers = [
            ('bit', BitType),
            ('tinyint', TinyIntType),
            ('smallint', SmallIntType),
            ('(?:int|integer)', IntType),
            ('bigint', BigIntType),
            ('real', RealType),
            ('(?:float|double precision)', FloatType),
            ('(?:char|character)', CharType),
            (r'(?:char|character)\((\d+)\)', lambda size_str: CharType(size=int(size_str))),
            (r'(?:varchar|char(?:|acter)\s+varying)', VarCharType),
            (r'(?:varchar|char(?:|acter)\s+varying)\((\d+)\)', lambda size_str: VarCharType(size=int(size_str))),
            (r'varchar\(max\)', VarCharMaxType),
            (r'(?:nchar|national\s+(?:char|character))', NCharType),
            (r'(?:nchar|national\s+(?:char|character))\((\d+)\)', lambda size_str: NCharType(size=int(size_str))),
            (r'(?:nvarchar|national\s+(?:char|character)\s+varying)', NVarCharType),
            (r'(?:nvarchar|national\s+(?:char|character)\s+varying)\((\d+)\)',
             lambda size_str: NVarCharType(size=int(size_str))),
            (r'nvarchar\(max\)', NVarCharMaxType),
            ('xml', XmlType),
            ('text', TextType),
            (r'(?:ntext|national\s+text)', NTextType),
            ('binary', BinaryType),
            ('binary\((\d+)\)', lambda size_str: BinaryType(size=int(size_str))),
            ('(?:varbinary|binary varying)', VarBinaryType),
            (r'(?:varbinary|binary varying)\((\d+)\)', lambda size_str: VarBinaryType(size=int(size_str))),
            (r'varbinary\(max\)', VarBinaryMaxType),
            ('image', ImageType),
            ('smalldatetime', SmallDateTimeType),
            ('datetime', DateTimeType),
            ('date', DateType),
            (r'time', TimeType),
            (r'time\((\d+)\)', lambda precision_str: TimeType(precision=int(precision_str))),
            ('datetime2', DateTime2Type),
            (r'datetime2\((\d+)\)', lambda precision_str: DateTime2Type(precision=int(precision_str))),
            ('datetimeoffset', DateTimeOffsetType),
            (r'datetimeoffset\((\d+)\)',
             lambda precision_str: DateTimeOffsetType(precision=int(precision_str))),
            ('(?:decimal|dec|numeric)', DecimalType),
            ('(?:decimal|dec|numeric)\((\d+)\)',
             lambda precision_str: DecimalType(precision=int(precision_str))),
            ('(?:decimal|dec|numeric)\((\d+), (\d+)\)',
             lambda precision_str, scale_str: DecimalType(precision=int(precision_str), scale=int(scale_str))),
            ('smallmoney', SmallMoneyType),
            ('money', MoneyType),
            ('uniqueidentifier', UniqueIdentifierType),
            ('sql_variant', VariantType),
        ]
        self._compiled = [(re.compile(r'^' + regex + '$', re.IGNORECASE), constructor)
                          for regex, constructor in declaration_parsers]

    def parse(self, declaration):
        """
        Parse sql type declaration, e.g. varchar(10) and return instance of corresponding type class,
        e.g. VarCharType(10)

        @param declaration: Sql declaration to parse, e.g. varchar(10)
        @return: instance of SqlTypeMetaclass
        """
        declaration = declaration.strip()
        for regex, constructor in self._compiled:
            m = regex.match(declaration)
            if m:
                return constructor(*m.groups())
        raise ValueError('Unable to parse type declaration', declaration)


_declarations_parser = DeclarationsParser()


class TdsTypeInferrer(object):
    def __init__(self, type_factory, collation=None, bytes_to_unicode=False, allow_tz=False):
        """
        Class used to do TDS type inference

        :param type_factory: Instance of TypeFactory
        :param collation: Collation to use for strings
        :param bytes_to_unicode: Treat bytes type as unicode string
        :param allow_tz: Allow usage of DATETIMEOFFSET type
        """
        self._type_factory = type_factory
        self._collation = collation
        self._bytes_to_unicode = bytes_to_unicode
        self._allow_tz = allow_tz

    def from_value(self, value):
        """ Function infers TDS type from Python value.

        :param value: value from which to infer TDS type
        :return: An instance of subclass of :class:`BaseType`
        """
        if value is None:
            sql_type = NVarCharType(size=1)
        else:
            sql_type = self._from_class_value(value, type(value))
        return sql_type

    def from_class(self, cls):
        """ Function infers TDS type from Python class.

        :param cls: Class from which to infer type
        :return: An instance of subclass of :class:`BaseType`
        """
        return self._from_class_value(None, cls)

    def _from_class_value(self, value, value_type):
        type_factory = self._type_factory
        bytes_to_unicode = self._bytes_to_unicode
        allow_tz = self._allow_tz

        if issubclass(value_type, bool):
            return BitType()
        elif issubclass(value_type, six.integer_types):
            if value is None:
                return IntType()
            if -2 ** 31 <= value <= 2 ** 31 - 1:
                return IntType()
            elif -2 ** 63 <= value <= 2 ** 63 - 1:
                return BigIntType()
            elif -10 ** 38 + 1 <= value <= 10 ** 38 - 1:
                return DecimalType(precision=38)
            else:
                return VarCharMaxType()
        elif issubclass(value_type, float):
            return FloatType()
        elif issubclass(value_type, Binary):
            if value is None or len(value) <= 8000:
                return VarBinaryType(size=8000)
            else:
                return type_factory.long_binary_type()
        elif issubclass(value_type, six.binary_type):
            if bytes_to_unicode:
                return type_factory.long_string_type()
            else:
                return type_factory.long_varchar_type()
        elif issubclass(value_type, six.string_types):
            return type_factory.long_string_type()
        elif issubclass(value_type, datetime.datetime):
            if value and value.tzinfo and allow_tz:
                return type_factory.datetime_with_tz(precision=6)
            else:
                return type_factory.datetime(precision=6)
        elif issubclass(value_type, datetime.date):
            return type_factory.date()
        elif issubclass(value_type, datetime.time):
            return type_factory.time(precision=6)
        elif issubclass(value_type, decimal.Decimal):
            if value is None:
                return DecimalType()
            else:
                return DecimalType.from_value(value)
        elif issubclass(value_type, uuid.UUID):
            return UniqueIdentifierType()
        elif issubclass(value_type, TableValuedParam):
            columns = value.columns
            rows = value.rows
            if columns is None:
                # trying to auto detect columns using data from first row
                if rows is None:
                    # rows are not present too, this means
                    # entire tvp has value of NULL
                    pass
                else:
                    # use first row to infer types of columns
                    row = value.peek_row()
                    columns = []
                    try:
                        cell_iter = iter(row)
                    except TypeError:
                        raise tds_base.DataError('Each row in table should be an iterable')
                    for cell in cell_iter:
                        if isinstance(cell, TableValuedParam):
                            raise tds_base.DataError('TVP type cannot have nested TVP types')
                        col_type = self.from_value(cell)
                        col = tds_base.Column(type=col_type)
                        columns.append(col)

            return TableType(typ_schema=value.typ_schema, typ_name=value.typ_name, columns=columns)
        else:
            raise tds_base.DataError('Cannot infer TDS type from Python value: {!r}'.format(value))



import socket
import socketserver
import struct
import logging

import OpenSSL.SSL

import pytds.tds


_BYTE_STRUCT = struct.Struct('B')
_OFF_LEN_STRUCT = struct.Struct('>HH')
_PROD_VER_STRUCT = struct.Struct('>LH')
logger = logging.getLogger(__name__)


class TdsParser:
    def bad_stream(self, msg):
        # TODO use different exception class
        raise Exception(msg)

    def parse_prelogin(self, buf):
        # https://msdn.microsoft.com/en-us/library/dd357559.aspx
        size = len(buf)
        i = 0
        result = {}
        while True:
            value = None
            if i >= size:
                self.bad_stream('Invalid size of PRELOGIN structure')
            type_id, = _BYTE_STRUCT.unpack_from(buf, i)
            if type_id == pytds.tds_base.PreLoginToken.TERMINATOR:
                break
            if i + 4 > size:
                self.bad_stream('Invalid size of PRELOGIN structure')
            off, l = _OFF_LEN_STRUCT.unpack_from(buf, i + 1)
            if off > size or off + l > size:
                self.bad_stream('Invalid offset in PRELOGIN structure')
            if type_id == pytds.tds_base.PreLoginToken.VERSION:
                value = _PROD_VER_STRUCT.unpack_from(buf, off)
            elif type_id == pytds.tds_base.PreLoginToken.ENCRYPTION:
                value = _BYTE_STRUCT.unpack_from(buf, off)[0]
            elif type_id == pytds.tds_base.PreLoginToken.MARS:
                value = bool(_BYTE_STRUCT.unpack_from(buf, off)[0])
            elif type_id == pytds.tds_base.PreLoginToken.INSTOPT:
                value = buf[off:off+l].decode('ascii')
            i += 5
            result[type_id] = value
        return result


class TdsGenerator:
    def generate_prelogin(self, prelogin):
        hdr_size = (1 + _OFF_LEN_STRUCT.size) * len(prelogin) + 1
        buf = bytearray([0] * hdr_size)
        hdr_offset = 0
        data_offset = hdr_size
        for type_id, value in prelogin.items():
            if type_id == pytds.tds_base.PreLoginToken.VERSION:
                packed = _PROD_VER_STRUCT.pack(value)
            elif type_id == pytds.tds_base.PreLoginToken.ENCRYPTION:
                packed = [value]
            elif type_id == pytds.tds_base.PreLoginToken.MARS:
                packed = [1 if value else 0]
            elif type_id == pytds.tds_base.PreLoginToken.INSTOPT:
                packed = value.encode('ascii')
            else:
                raise Exception(f"not implemented prelogin option {type_id} in prelogin message generator")

            data_size = len(packed)

            buf[hdr_offset] = type_id
            hdr_offset += 1

            _OFF_LEN_STRUCT.pack_into(buf, hdr_offset, data_offset, data_size)
            hdr_offset += _OFF_LEN_STRUCT.size

            buf.extend(packed)
            data_offset += data_size
        buf[hdr_offset] = pytds.tds_base.PreLoginToken.TERMINATOR
        return buf


class Sock():
    # wraps request in class compatible with TdsSocket
    def __init__(self, req):
        self._req = req

    def recv(self, size):
        return self._req.recv(size)

    def recv_into(self, buffer, size=0):
        return self._req.recv_into(buffer, size)

    def sendall(self, data, flags=0):
        return self._req.sendall(data, flags)


class RequestHandler(socketserver.StreamRequestHandler):
    def handle(self):
        parser = TdsParser()
        gen = TdsGenerator()
        bufsize = 4096

        # TdsReader expects this
        self._transport = Sock(self.request)

        r = pytds.tds._TdsReader(self)
        w = pytds.tds._TdsWriter(self, bufsize=bufsize)

        buf = r.read_whole_packet()
        if r.packet_type != pytds.tds_base.PacketType.PRELOGIN:
            msg = 'Invalid packet type: {0}, expected PRELOGIN({1})'.format(r.packet_type,
                                                                            pytds.tds_base.PacketType.PRELOGIN)
            self.bad_stream(msg)
        prelogin = parser.parse_prelogin(buf)
        logger.info(f"received prelogin message from client {prelogin}")
        srv_enc = self.server._enc
        cli_enc = prelogin[pytds.tds_base.PreLoginToken.ENCRYPTION]
        res_enc = None
        close_conn = False
        if srv_enc == pytds.PreLoginEnc.ENCRYPT_OFF:
            if cli_enc == pytds.PreLoginEnc.ENCRYPT_OFF:
                res_enc = pytds.PreLoginEnc.ENCRYPT_OFF
            elif cli_enc == pytds.PreLoginEnc.ENCRYPT_ON:
                res_enc = pytds.PreLoginEnc.ENCRYPT_ON
            elif cli_enc == pytds.PreLoginEnc.ENCRYPT_NOT_SUP:
                res_enc = pytds.PreLoginEnc.ENCRYPT_NOT_SUP
        elif srv_enc == pytds.PreLoginEnc.ENCRYPT_ON:
            if cli_enc == pytds.PreLoginEnc.ENCRYPT_OFF:
                res_enc = pytds.PreLoginEnc.ENCRYPT_REQ
            elif cli_enc == pytds.PreLoginEnc.ENCRYPT_ON:
                res_enc = pytds.PreLoginEnc.ENCRYPT_ON
            elif cli_enc == pytds.PreLoginEnc.ENCRYPT_NOT_SUP:
                res_enc = pytds.PreLoginEnc.ENCRYPT_REQ
                close_conn = True
        elif srv_enc == pytds.PreLoginEnc.ENCRYPT_NOT_SUP:
            if cli_enc == pytds.PreLoginEnc.ENCRYPT_OFF:
                res_enc = pytds.PreLoginEnc.ENCRYPT_NOT_SUP
            elif cli_enc == pytds.PreLoginEnc.ENCRYPT_ON:
                res_enc = pytds.PreLoginEnc.ENCRYPT_NOT_SUP
                close_conn = True
            elif cli_enc == pytds.PreLoginEnc.ENCRYPT_NOT_SUP:
                res_enc = pytds.PreLoginEnc.ENCRYPT_NOT_SUP
        elif srv_enc == pytds.PreLoginEnc.ENCRYPT_REQ:
            res_enc = pytds.PreLoginEnc.ENCRYPT_REQ

        # sending reply to client's prelogin packet
        prelogin_resp = gen.generate_prelogin({
            pytds.tds_base.PreLoginToken.ENCRYPTION: res_enc,
        })
        w.begin_packet(pytds.tds_base.PacketType.REPLY)
        w.write(prelogin_resp)
        w.flush()

        if close_conn:
            return

        wrapped_socket = None
        if res_enc != pytds.PreLoginEnc.ENCRYPT_NOT_SUP:
            # setup TLS connection
            tlsconn = OpenSSL.SSL.Connection(self.server._tls_ctx)
            tlsconn.set_accept_state()
            done = False
            while not done:
                try:
                    tlsconn.do_handshake()
                except OpenSSL.SSL.WantReadError:
                    try:
                        buf = tlsconn.bio_read(bufsize)
                    except OpenSSL.SSL.WantReadError:
                        pass
                    else:
                        w.begin_packet(pytds.tds_base.PacketType.REPLY)
                        w.write(buf)
                        w.flush()

                    buf = r.read_whole_packet()
                    tlsconn.bio_write(buf)
                else:
                    done = True
                    try:
                        buf = tlsconn.bio_read(bufsize)
                    except OpenSSL.SSL.WantReadError:
                        pass
                    else:
                        w.begin_packet(pytds.tds_base.PacketType.REPLY)
                        w.write(buf)
                        w.flush()

            wrapped_socket = pytds.tls.EncryptedSocket(transport=self.request, tls_conn=tlsconn)
            r._transport = wrapped_socket
            w._transport = wrapped_socket

        try:
            buf = r.read_whole_packet()
        except pytds.tds_base.ClosedConnectionError:
            logger.info('client closed connection, probably did not like server certificate')
            return
        logger.info(f"received login packet from client {buf}")

        if res_enc == pytds.PreLoginEnc.ENCRYPT_OFF:
            wrapped_socket.shutdown()
            r._transport = self._transport
            w._transport = self._transport

        srv_name = 'Simple TDS Server'
        srv_ver = (1, 0, 0, 0)
        tds_version = self.server._tds_version

        w.begin_packet(pytds.tds_base.PacketType.REPLY)
        # https://msdn.microsoft.com/en-us/library/dd340651.aspx
        srv_name_coded, _ = pytds.tds.ucs2_codec.encode(srv_name)
        srv_name_size = len(srv_name_coded)
        w.put_byte(pytds.tds_base.TDS_LOGINACK_TOKEN)
        size = 1 + 4 + 1 + srv_name_size + 4
        w.put_usmallint(size)
        w.put_byte(1)  # interface
        w.put_uint_be(tds_version)
        w.put_byte(len(srv_name))
        w.write(srv_name_coded)
        w.put_byte(srv_ver[0])
        w.put_byte(srv_ver[1])
        w.put_byte(srv_ver[2])
        w.put_byte(srv_ver[3])

        # https://msdn.microsoft.com/en-us/library/dd340421.aspx
        w.put_byte(pytds.tds_base.TDS_DONE_TOKEN)
        w.put_usmallint(0)  # status
        w.put_usmallint(0)  # curcmd
        w.put_uint8(0)   # done row count

        w.flush()

    def bad_stream(self, msg):
        raise Exception(msg)


class SimpleServer(socketserver.TCPServer):
    allow_reuse_address = True

    def __init__(self, address, enc, cert=None, pkey=None, tds_version=pytds.tds_base.TDS74):
        self._enc = enc
        super().__init__(address, RequestHandler)
        ctx = None
        if cert and pkey:
            ctx = OpenSSL.SSL.Context(OpenSSL.SSL.TLSv1_2_METHOD)
            ctx.set_options(OpenSSL.SSL.OP_NO_SSLv2)
            ctx.set_options(OpenSSL.SSL.OP_NO_SSLv3)
            ctx.use_certificate(cert)
            ctx.use_privatekey(pkey)
        self._tls_ctx = ctx
        self._tds_version = tds_version

    def set_ssl_context(self, ctx):
        self._tls_ctx = ctx

    def set_enc(self, enc):
        self._enc = enc


def run(address):
    logger.info('Starting server...')
    with SimpleServer(address) as server:
        logger.info('Press Ctrl+C to stop the server')
        server.serve_forever()


if __name__ == '__main__':
    run()
